from __future__ import annotations
from typing import List
import labelme
import json
from tqdm import tqdm

from logger import logger
from common_utils.common_types.point import Point2D_List
from common_utils.check_utils import check_required_keys, check_type, check_file_exists, check_value, \
    check_dir_exists
from common_utils.path_utils import get_all_files_of_extension, get_rootname_from_path, get_filename, get_abs_path, \
    get_dirpath_from_filepath
from common_utils.file_utils import delete_all_files_in_dir, make_dir_if_not_exists, file_exists, copy_file

class LabelmeShape:
    def __init__(
        self,
        label: str, points: Point2D_List, shape_type: str,
        group_id: int=None, flags: dict={}
    ):
        self._check_valid(shape_type=shape_type, points=points)
        self.label = label
        self.points = points
        self.shape_type = shape_type
        self.group_id = group_id
        self.flags = flags

    @staticmethod
    def _check_valid(shape_type: str, points: Point2D_List):
        check_value(shape_type, valid_value_list=['polygon', 'rectangle', 'circle', 'line', 'point', 'linestrip'])
        if shape_type == 'polygon':
            if len(points) < 3:
                logger.error(f'Labelme polygon requires at least 3 points.')
                raise Exception
        elif shape_type == 'rectangle':
            if len(points) != 2:
                logger.error(f'Labelme rectangle requires exactly 2 points.')
                raise Exception
        elif shape_type == 'circle':
            if len(points) != 2:
                logger.error(f'Labelme circle requires exactly 2 points.')
                raise Exception
        elif shape_type == 'line':
            if len(points) != 2:
                logger.error(f'Labelme line requires exactly 2 points.')
                raise Exception
        elif shape_type == 'point':
            if len(points) != 1:
                logger.error(f'Labelme point requires exactly 1 points.')
                raise Exception
        elif shape_type == 'linestrip':
            if len(points) < 2:
                logger.error(f'Labelme linestrip requires at least 2 points.')
                raise Exception

    def to_dict(self) -> dict:
        return {
            'label': self.label,
            'points': self.points.to_list(demarcation=True),
            'shape_type': self.shape_type,
            'group_id': self.group_id,
            'flags': self.flags
        }

    @classmethod
    def from_dict(cls, shape_dict: dict) -> LabelmeShape:
        check_required_keys(shape_dict, required_keys=['label', 'points', 'group_id', 'shape_type', 'flags'])
        return LabelmeShape(
            label=shape_dict['label'],
            points=Point2D_List.from_list(value_list=shape_dict['points'], demarcation=True),
            shape_type=shape_dict['shape_type'],
            group_id=shape_dict['group_id'],
            flags=shape_dict['flags']
        )

class LabelmeShapeHandler:
    def __init__(self, shape_list: List[LabelmeShape]=None):
        self.shape_list = shape_list if shape_list is not None else []

    def __len__(self) -> int:
        return len(self.shape_list)

    def __getitem__(self, idx: int) -> LabelmeShape:
        if len(self.shape_list) == 0:
            logger.error(f"LabelmeShapeHandler is empty.")
            raise IndexError
        elif idx < 0 or idx >= len(self.shape_list):
            logger.error(f"Index out of range: {idx}")
            raise IndexError
        else:
            return self.shape_list[idx]

    def __setitem__(self, idx: int, value: LabelmeShape):
        check_type(value, valid_type_list=[LabelmeShape])
        self.shape_list[idx] = value

    def __iter__(self):
        self.n = 0
        return self

    def __next__(self) -> LabelmeShape:
        if self.n < len(self.shape_list):
            result = self.shape_list[self.n]
            self.n += 1
            return result
        else:
            raise StopIteration

    def append(self, shape: LabelmeShape):
        check_type(shape, valid_type_list=[LabelmeShape])
        self.shape_list.append(shape)

    def to_dict_list(self) -> List[dict]:
        return [shape.to_dict() for shape in self]

    @classmethod
    def from_dict_list(cls, dict_list: List[dict]) -> LabelmeShapeHandler:
        return LabelmeShapeHandler(
            shape_list=[LabelmeShape.from_dict(shape_dict) for shape_dict in dict_list]
        )

class LabelmeAnnotation:
    def __init__(
        self,
        img_path: str, img_h: int, img_w: int,
        version: str=labelme.__version__, flags: dict={},
        shapes: LabelmeShapeHandler=None,
        img_data: str=None
    ):
        self.img_path = img_path
        self.img_h, self.img_w = img_h, img_w
        self.version = version
        self.flags = flags
        self.shapes = shapes if shapes is not None else LabelmeShapeHandler()
        self.img_data = img_data

    def to_dict(self) -> dict:
        return {
            'version': self.version,
            'flags': self.flags,
            'shapes': self.shapes.to_dict_list(),
            'imagePath': self.img_path,
            'imageData': self.img_data,
            'imageHeight': self.img_h,
            'imageWidth': self.img_w
        }
    
    @classmethod
    def from_dict(cls, labelme_dict: dict) -> LabelmeAnnotation:
        check_required_keys(labelme_dict, required_keys=['version', 'flags', 'shapes', 'imagePath', 'imageData', 'imageHeight', 'imageWidth'])
        return LabelmeAnnotation(
            version=labelme_dict['version'],
            flags=labelme_dict['flags'],
            shapes=LabelmeShapeHandler.from_dict_list(labelme_dict['shapes']),
            img_path=labelme_dict['imagePath'],
            img_data=labelme_dict['imageData'],
            img_h=labelme_dict['imageHeight'],
            img_w=labelme_dict['imageWidth']
        )

    def save_to_path(self, save_path: str, overwrite: bool=False, img_path: str=None):
        if file_exists(save_path) and not overwrite:
            logger.error(f'File already exists at save_path: {save_path}')
            raise Exception
        if img_path is not None:
            import os
            self.img_path = os.path.relpath(path=img_path, start=get_dirpath_from_filepath(save_path))
        json_dict = self.to_dict()
        json.dump(json_dict, open(save_path, 'w'), indent=2, ensure_ascii=False)

    @classmethod
    def load_from_path(cls, json_path: str) -> LabelmeAnnotation:
        check_file_exists(json_path)
        json_dict = json.load(open(json_path, 'r'))
        return LabelmeAnnotation.from_dict(json_dict)

class LabelmeAnnotationHandler:
    def __init__(self, labelme_ann_list: List[LabelmeAnnotation]=None):
        self.labelme_ann_list = labelme_ann_list if labelme_ann_list is not None else []

    def __len__(self) -> int:
        return len(self.labelme_ann_list)

    def __getitem__(self, idx: int) -> LabelmeAnnotation:
        if len(self.labelme_ann_list) == 0:
            logger.error(f"LabelmeAnnotationHandler is empty.")
            raise IndexError
        elif idx < 0 or idx >= len(self.labelme_ann_list):
            logger.error(f"Index out of range: {idx}")
            raise IndexError
        else:
            return self.labelme_ann_list[idx]

    def __setitem__(self, idx: int, value: LabelmeAnnotation):
        check_type(value, valid_type_list=[LabelmeAnnotation])
        self.labelme_ann_list[idx] = value

    def __iter__(self):
        self.n = 0
        return self

    def __next__(self) -> LabelmeAnnotation:
        if self.n < len(self.labelme_ann_list):
            result = self.labelme_ann_list[self.n]
            self.n += 1
            return result
        else:
            raise StopIteration

    def append(self, ann: LabelmeAnnotation):
        check_type(ann, valid_type_list=[LabelmeAnnotation])
        self.labelme_ann_list.append(ann)

    def erase_image_data(self):
        for ann in self:
            ann.img_data = None

    def _check_paths_valid(self, src_img_dir: str):
        check_dir_exists(src_img_dir)
        img_filename_list = []
        duplicate_img_filename_list = []
        for ann in self:
            img_filename = get_filename(ann.img_path)
            if img_filename not in img_filename_list:
                img_filename_list.append(ann.img_path)
            else:
                duplicate_img_filename_list.append(ann.img_path)
            img_path = f'{src_img_dir}/{img_filename}'
            check_file_exists(img_path)
        if len(duplicate_img_filename_list) > 0:
            logger.error(f'Found the following duplicate image filenames in LabelmeAnnotationHandler:\n{duplicate_img_filename_list}')
            raise Exception

    def save_to_dir(self, json_save_dir: str, src_img_dir: str, overwrite: bool=False, dst_img_dir: str=None):
        self._check_paths_valid(src_img_dir=src_img_dir)
        make_dir_if_not_exists(json_save_dir)
        delete_all_files_in_dir(json_save_dir, ask_permission=not overwrite)
        if dst_img_dir is not None:
            make_dir_if_not_exists(dst_img_dir)
            delete_all_files_in_dir(dst_img_dir, ask_permission=not overwrite)

        for ann in tqdm(self, total=len(self), unit='ann', leave=True):
            save_path = f'{json_save_dir}/{get_rootname_from_path(ann.img_path)}.json'
            src_img_path = f'{src_img_dir}/{get_filename(ann.img_path)}'
            if dst_img_dir is not None:
                dst_img_path = f'{dst_img_dir}/{get_filename(ann.img_path)}'
                copy_file(src_path=src_img_path, dest_path=dst_img_path, silent=True)
                ann.save_to_path(save_path=save_path, img_path=dst_img_path)
            else:
                ann.save_to_path(save_path=save_path, img_path=src_img_path)

    @classmethod
    def load_from_pathlist(cls, json_path_list: list) -> LabelmeAnnotationHandler:
        return LabelmeAnnotationHandler(
            labelme_ann_list=[LabelmeAnnotation.load_from_path(json_path) for json_path in json_path_list]
        )

    @classmethod
    def load_from_dir(cls, load_dir: str) -> LabelmeAnnotationHandler:
        check_dir_exists(load_dir)
        json_path_list = get_all_files_of_extension(dir_path=load_dir, extension='json')
        return cls.load_from_pathlist(json_path_list)


from __future__ import annotations
from typing import List
from logger import logger
from common_utils.check_utils import check_type_from_list, check_type
from common_utils.common_types.bbox import BBox
from common_utils.common_types.segmentation import Polygon
from common_utils.common_types.keypoint import Keypoint2D
from .objects import COCO_Category

class KeypointGroup:
    def __init__(
        self, bound_obj, coco_cat: COCO_Category,
        kpt_list: List[Keypoint2D]=None, kpt_label_list: List[str]=None,
        img_filename: str=None
    ):
        check_type(bound_obj, valid_type_list=[BBox, Polygon])
        self.bound_obj = bound_obj
        check_type(coco_cat, valid_type_list=[COCO_Category])
        self.coco_cat = coco_cat
        if kpt_list is not None:
            check_type_from_list(kpt_list, valid_type_list=[Keypoint2D])
            self.kpt_list = kpt_list
        else:
            self.kpt_list = []
        self.kpt_label_list = []
        self.img_filename = img_filename
        if kpt_list is not None or kpt_label_list is not None:
            if kpt_list is None or kpt_label_list is None:
                logger.error(f'Must provide both kpt_list and kpt_label_list, or neither.')
                logger.error(f'kpt_list: {kpt_list}\nkpt_label_list: {kpt_label_list}')
                logger.error(f'Ground truth labels: {self.coco_cat.keypoints}')
                if self.img_filename is not None:
                    logger.error(f'Image filename: {self.img_filename}')
                raise Exception
            if len(kpt_list) != len(kpt_label_list):
                logger.error(f'len(kpt_list) == {len(kpt_list)} != {len(kpt_label_list)} == len(kpt_label_list)')
                raise Exception
            for kpt, label in zip(kpt_list, kpt_label_list):
                self.register(kpt=kpt, label=label)
        self.postponed_kpt_list = []
        self.postponed_kpt_label_list = []
    
    def register(self, kpt: Keypoint2D, label: str, strict: bool=True):
        if label in self.kpt_label_list:
            logger.error(f"Keypoint label '{label}' already exists in self.kpt_list.")
            logger.error(f'Currently registered labels: {self.kpt_label_list}')
            logger.error(f'Ground truth labels: {self.coco_cat.keypoints}')
            if self.img_filename is not None:
                logger.error(f'Image filename: {self.img_filename}')
            raise Exception
        if label not in self.coco_cat.keypoints:
            if not strict:
                raise NotImplementedError
                self.postponed_kpt_list.append(kpt)
                self.postponed_kpt_label_list.append(label)
            else:
                logger.error(f"Label {label} couldn't be found in corresponding category's keypoint list.")
                logger.error(f'coco_cat.keypoints: {self.coco_cat.keypoints}')
                raise Exception
        else:
            self.kpt_list.append(kpt)
            self.kpt_label_list.append(label)


from __future__ import annotations
from abc import ABCMeta, abstractmethod
from typing import TypeVar, Type, Generic, List
import json
import operator
import random

from logger import logger
from common_utils.check_utils import check_required_keys, check_type_from_list, \
    check_type
from common_utils.file_utils import file_exists

T = TypeVar('T')
H = TypeVar('H')

class BaseStructObject(Generic[T]):
    def __init__(self):
        pass

    @abstractmethod
    def __str__(self) -> str:
        ''' To override '''
        raise NotImplementedError

    def __repr__(self):
        return self.__str__()

    def __key(self) -> tuple:
        return tuple([self.__class__] + list(self.__dict__.values()))

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other) -> bool:
        if isinstance(other, self.__class__):
            return self.__key() == other.__key()
        return NotImplemented

    @classmethod
    def buffer(cls: T, obj) -> T:
        return obj

    def copy(self: T) -> T:
        """Note: All class variables must be part of the constructor."""
        return type(self)(*self.__dict__.values())

    def to_dict(self: T) -> dict:
        """Note: All class variables will be put into the dict."""
        return self.__dict__

    def save_to_path(self: T, save_path: str, overwrite: bool=False):
        if file_exists(save_path) and not overwrite:
            logger.error(f'File already exists at save_path: {save_path}')
            raise Exception
        json_dict = self.to_dict()
        json.dump(json_dict, open(save_path, 'w'), indent=2, ensure_ascii=False)

class BaseStructHandler(Generic[H, T]):
    def __init__(self: H, obj_type: type, obj_list: List[T]=None):
        check_type(obj_type, valid_type_list=[type])
        self.obj_type = obj_type
        if obj_list is not None:
            check_type_from_list(obj_list, valid_type_list=[obj_type])
        self.obj_list = obj_list if obj_list is not None else []

    def __str__(self):
        print_str = ""
        for obj in self.obj_list:
            print_str += f"{obj}\n"

        return print_str

    def __repr__(self):
        return self.__str__()

    def __len__(self) -> int:
        return len(self.obj_list)

    def __getitem__(self, idx: int) -> T:
        if type(idx) is int:
            if len(self.obj_list) == 0:
                logger.error(f"{type(self).__name__} is empty.")
                raise IndexError
            elif idx < 0 or idx >= len(self.obj_list):
                logger.error(f"Index out of range: {idx}")
                raise IndexError
            else:
                return self.obj_list[idx]
        elif type(idx) is slice:
            return self.obj_list[idx.start:idx.stop:idx.step]
        else:
            logger.error(f'Expected int or slice. Got type(idx)={type(idx)}')
            raise TypeError

    def __setitem__(self, idx: int, value: T):
        check_type(value, valid_type_list=[self.obj_type])
        if type(idx) is int:
            self.obj_list[idx] = value
        elif type(idx) is slice:
            self.obj_list[idx.start:idx.stop:idx.step] = value
        else:
            logger.error(f'Expected int or slice. Got type(idx)={type(idx)}')
            raise TypeError

    def __delitem__(self, idx):
        if type(idx) is int:
            if len(self.obj_list) == 0:
                logger.error(f"{type(self).__name__} is empty.")
                raise IndexError
            elif idx < 0 or idx >= len(self.obj_list):
                logger.error(f"Index out of range: {idx}")
                raise IndexError
            else:
                del self.obj_list[idx]
        elif type(idx) is slice:
            del self.obj_list[idx.start:idx.stop:idx.step]
        else:
            logger.error(f'Expected int or slice. Got type(idx)={type(idx)}')
            raise TypeError

    def __iter__(self):
        self.n = 0
        return self

    def __next__(self) -> T:
        if self.n < len(self.obj_list):
            result = self.obj_list[self.n]
            self.n += 1
            return result
        else:
            raise StopIteration

    def copy(self) -> H:
        return type(self)(self.obj_list.copy())

    def append(self, item: T):
        check_type(item, valid_type_list=[self.obj_type])
        self.obj_list.append(item)

    def sort(self, attr_name: str, reverse: bool=False):
        if len(self) > 0:
            attr_list = list(self.obj_list[0].__dict__.keys())    
            if attr_name not in attr_list:
                logger.error(f"{self.obj_type.__name__} class has not attribute: '{attr_name}'")
                logger.error(f'Possible attribute names:')
                for name in attr_list:
                    logger.error(f'\t{name}')
                raise Exception

            self.obj_list.sort(key=operator.attrgetter(attr_name), reverse=reverse)
        else:
            logger.error(f"Cannot sort. {type(self).__name__} is empty.")
            raise Exception

    def shuffle(self):
        random.shuffle(self.obj_list)

    def get_obj_from_id(self, id: int) -> T:
        id_list = []
        for obj in self:
            if id == obj.id:
                return obj
            else:
                id_list.append(obj.id)
        id_list.sort()
        logger.error(f"Couldn't find {self.obj_type.__name__} with id={id}")
        logger.error(f"Possible ids: {id_list}")
        raise Exception

    def to_dict_list(self) -> List[dict]:
        return [item.to_dict() for item in self]

    def save_to_path(self, save_path: str, overwrite: bool=False):
        if file_exists(save_path) and not overwrite:
            logger.error(f'File already exists at save_path: {save_path}')
            raise Exception
        json_data = self.to_dict_list()
        json.dump(json_data, open(save_path, 'w'), indent=2, ensure_ascii=False)


from __future__ import annotations
from typing import List
import yaml
import json

from logger import logger
from common_utils.file_utils import file_exists
from common_utils.path_utils import get_extension_from_path, rel_to_abs_path
from common_utils.check_utils import check_required_keys, check_type, \
    check_type_from_list, check_list_length, check_file_exists, \
    check_dir_exists

from ....base import BaseStructObject, BaseStructHandler
from .path import Path

class DatasetConfig(BaseStructObject['DatasetConfig']):
    def __init__(self, img_dir: str, ann_path: str, ann_format: str='coco', tag: str=None):
        super().__init__()
        self.img_dir = img_dir
        self.ann_path = ann_path
        self.ann_format = ann_format
        self.tag = tag

    def __str__(self) -> str:
        return str(self.__dict__)

class DatasetConfigCollection(BaseStructHandler['DatasetConfigCollection', 'DatasetConfig']):
    def __init__(self, dataset_config_list: List[DatasetConfig]=None, tag: str=None):
        super().__init__(obj_type=DatasetConfig, obj_list=dataset_config_list)
        self.dataset_config_list = self.obj_list
        self.tag = tag

    def to_dict0(self) -> dict:
        # TODO: Work In Progress
        img_dir_list = [Path(config.img_dir).abs() for config in self.dataset_config_list]
        ann_path_list = [Path(config.ann_path).abs() for config in self.dataset_config_list]
        ann_format_list = [config.ann_format for config in self.dataset_config_list]
        img_container_dir = Path.get_longest_container_dir(img_dir_list)
        ann_container_dir = Path.get_longest_container_dir(ann_path_list)
        collection_dir = Path.get_longest_container_dir([img_container_dir, ann_container_dir])
        dataset_names = [
            img_dir.replace(f'{collection_dir.path_str}/', '') \
                if img_dir != collection_dir else Path('') \
                for img_dir in img_dir_list
        ]
        
        # rel_img_dir_list = [None] * len(dataset_names)
        rel_img_dir_list = [Path('') * len(dataset_names)]
        
        logger.purple(f'Before dataset_names: {dataset_names}')
        logger.purple(f'Before rel_img_dir_list: {rel_img_dir_list}')

        while True:
            logger.blue(f'Flag0')
            # Adjust dataset_names tails
            tail_moved = Path.tail2head(dataset_names, rel_img_dir_list)

            if not tail_moved:
                break
        
        while True:
            logger.blue(f'Flag1')
            # Adjust dataset_names heads
            head_moved = Path.head2tail(dataset_names, collection_dir)

            if not head_moved:
                break

        logger.purple(f'After dataset_names: {dataset_names}')
        logger.purple(f'After rel_img_dir_list: {rel_img_dir_list}')

        rel_img_dir_list = [rel_img_dir if rel_img_dir is not None else Path('') for rel_img_dir in rel_img_dir_list]
        rel_ann_path_list = [
            ann_path.replace(f'{collection_dir}/{dataset_name}/', '') \
                if dataset_name != Path('') else ann_path.replace(f'{collection_dir}/', '') \
                for ann_path, dataset_name in zip(ann_path_list, dataset_names)
        ]

        dataset_names = [dataset_name.path_str if dataset_name.path_str != '' else '.' for dataset_name in dataset_names]
        rel_img_dir = rel_img_dir_list[0].path_str if len(list(dict.fromkeys(rel_img_dir_list))) == 1 else [rel_img_dir.path_str for rel_img_dir in rel_img_dir_list]
        if type(rel_img_dir) is str:
            rel_img_dir = rel_img_dir if rel_img_dir != '' else '.'
        elif type(rel_img_dir) is list:
            rel_img_dir = [dir_path if dir_path != '' else '.' for dir_path in rel_img_dir]
        else:
            raise Exception
        rel_ann_path = rel_ann_path_list[0].path_str if len(list(dict.fromkeys(rel_ann_path_list))) == 1 else [rel_ann_path.path_str for rel_ann_path in rel_ann_path_list]
        ann_format = ann_format_list[0] if len(list(dict.fromkeys(ann_format_list))) == 1 else ann_format_list

        return {
            'collection_dir': collection_dir.path_str,
            'dataset_names': dataset_names,
            'dataset_specific': {
                'img_dir': rel_img_dir,
                'ann_path': rel_ann_path,
                'ann_format': ann_format
            }
        }

    def to_dict(self) -> dict:
        # TODO: May need some fixing
        if len(self.dataset_config_list) == 0:
            return {}
        img_dir_list = [Path(config.img_dir).abs() for config in self.dataset_config_list]
        ann_path_list = [Path(config.ann_path).abs() for config in self.dataset_config_list]
        ann_format_list = [config.ann_format for config in self.dataset_config_list]
        img_container_dir = Path.get_longest_container_dir(img_dir_list)
        ann_container_dir = Path.get_longest_container_dir(ann_path_list)
        collection_dir = Path.get_longest_container_dir([img_container_dir, ann_container_dir])
        dataset_names = [
            img_dir.replace(f'{collection_dir.path_str}/', '') \
                if img_dir != collection_dir else Path('') \
                for img_dir in img_dir_list
        ]
        
        rel_img_dir_list = [None] * len(dataset_names)
        
        done = False
        while not done:
            # Adjust dataset_names tails
            uniform_tail = None
            for i in range(len(dataset_names)):
                if len(dataset_names[i]) == 0:
                    continue
                for j in range(i+1, len(dataset_names)):
                    if len(dataset_names[j]) == 0:
                        continue
                    if dataset_names[i][-1] != dataset_names[j][-1]:
                        uniform_tail = False
                        break
                    elif uniform_tail is None:
                        uniform_tail = True
                if uniform_tail is not None and not uniform_tail:
                    break
            if uniform_tail:
                for i in range(len(dataset_names)):
                    if rel_img_dir_list[i] is None:
                        rel_img_dir_list[i] = dataset_names[i][-1]
                    else:
                        rel_img_dir_list[i] = dataset_names[i][-1] + rel_img_dir_list[i]
                    del dataset_names[i][-1]

            # Adjust dataset_names heads
            uniform_head = None
            for i in range(len(dataset_names)):
                if len(dataset_names[i]) == 0:
                    continue
                for j in range(i+1, len(dataset_names)):
                    if len(dataset_names[j]) == 0:
                        continue
                    if dataset_names[i][0] != dataset_names[j][0]:
                        uniform_head = False
                        break
                    elif uniform_head is None:
                        uniform_head = True
                if uniform_head is not None and not uniform_head:
                    break
            if uniform_head:
                collection_dir = collection_dir + dataset_names[0][0]
                for i in range(len(dataset_names)):
                    del dataset_names[i][0]
            
            if not uniform_tail and not uniform_head:
                done = True

        rel_img_dir_list = [rel_img_dir if rel_img_dir is not None else Path('') for rel_img_dir in rel_img_dir_list]
        rel_ann_path_list = [
            ann_path.replace(f'{collection_dir}/{dataset_name}/', '') \
                if dataset_name != Path('') else ann_path.replace(f'{collection_dir}/', '') \
                for ann_path, dataset_name in zip(ann_path_list, dataset_names)
        ]

        dataset_names = [dataset_name.path_str if dataset_name.path_str != '' else '.' for dataset_name in dataset_names]
        rel_img_dir = rel_img_dir_list[0].path_str if len(list(dict.fromkeys(rel_img_dir_list))) == 1 else [rel_img_dir.path_str for rel_img_dir in rel_img_dir_list]
        if type(rel_img_dir) is str:
            rel_img_dir = rel_img_dir if rel_img_dir != '' else '.'
        elif type(rel_img_dir) is list:
            rel_img_dir = [dir_path if dir_path != '' else '.' for dir_path in rel_img_dir]
        else:
            raise Exception
        rel_ann_path = rel_ann_path_list[0].path_str if len(list(dict.fromkeys(rel_ann_path_list))) == 1 else [rel_ann_path.path_str for rel_ann_path in rel_ann_path_list]
        ann_format = ann_format_list[0] if len(list(dict.fromkeys(ann_format_list))) == 1 else ann_format_list
        tag = [config.tag for config in self]
        tag = tag[0] if len(list(dict.fromkeys(tag))) == 1 else tag

        return {
            'collection_dir': collection_dir.path_str,
            'dataset_names': dataset_names,
            'dataset_specific': {
                'img_dir': rel_img_dir,
                'ann_path': rel_ann_path,
                'ann_format': ann_format,
                'tag': tag
            },
            'tag': self.tag
        }

    @classmethod
    def from_dict(cls, collection_dict: dict, check_paths: bool=True) -> DatasetConfigCollection:
        check_required_keys(
            collection_dict,
            required_keys=[
                'collection_dir', 'dataset_names', 'dataset_specific'
            ]
        )
        collection_dir = collection_dict['collection_dir']
        check_type(collection_dir, valid_type_list=[str])
        dataset_names = collection_dict['dataset_names']
        check_type(dataset_names, valid_type_list=[list])
        check_type_from_list(dataset_names, valid_type_list=[str])
        dataset_specific = collection_dict['dataset_specific']
        check_type(dataset_specific, valid_type_list=[dict])
        collection_tag = None if 'tag' not in collection_dict else collection_dict['tag']
        check_type(collection_tag, valid_type_list=[type(None), str])
        check_required_keys(
            dataset_specific,
            required_keys=[
                'img_dir', 'ann_path', 'ann_format'
            ]
        )
        img_dir = dataset_specific['img_dir']
        check_type(img_dir, valid_type_list=[str, list])
        if type(img_dir) is list:
            check_type_from_list(img_dir, valid_type_list=[str])
            check_list_length(img_dir, correct_length=len(dataset_names))
        ann_path = dataset_specific['ann_path']
        check_type(ann_path, valid_type_list=[str, list])
        if type(ann_path) is list:
            check_type_from_list(ann_path, valid_type_list=[str])
            check_list_length(ann_path, correct_length=len(dataset_names))
        ann_format = dataset_specific['ann_format']
        check_type(ann_format, valid_type_list=[str, list])
        if type(ann_format) is list:
            check_type_from_list(ann_format, valid_type_list=[str])
            check_list_length(ann_format, correct_length=len(dataset_names))
        dataset_tag = None if 'tag' not in dataset_specific else dataset_specific['tag']
        check_type(dataset_tag, valid_type_list=[type(None), str, list])
        if type(dataset_tag) is list:
            check_type_from_list(dataset_tag, valid_type_list=[type(None), str])
            check_list_length(dataset_tag, correct_length=len(dataset_names))

        dataset_config_list = []
        for i in range(len(dataset_names)):
            if type(img_dir) is str:
                img_dir0 = img_dir
            elif type(img_dir) is list:
                if i >= len(img_dir):
                    raise IndexError
                img_dir0 = img_dir[i]
            else:
                raise Exception

            if type(ann_path) is str:
                ann_path0 = ann_path
            elif type(ann_path) is list:
                if i >= len(ann_path):
                    raise IndexError
                ann_path0 = ann_path[i]
            else:
                raise Exception

            if type(ann_format) is str:
                ann_format0 = ann_format
            elif type(ann_format) is list:
                if i >= len(ann_format):
                    raise IndexError
                ann_format0 = ann_format[i]
            else:
                raise Exception

            if type(dataset_tag) is str or dataset_tag is None:
                dataset_tag0 = dataset_tag
            elif type(dataset_tag) is list:
                if i >= len(dataset_tag):
                    raise IndexError
                dataset_tag0 = dataset_tag[i]
            else:
                raise Exception

            img_dir1 = rel_to_abs_path(f'{collection_dir}/{dataset_names[i]}/{img_dir0}')
            ann_path1 = rel_to_abs_path(f'{collection_dir}/{dataset_names[i]}/{ann_path0}')
            if check_paths:
                check_dir_exists(img_dir1)
                check_file_exists(ann_path1)
            config = DatasetConfig(
                img_dir=img_dir1,
                ann_path=ann_path1,
                ann_format=ann_format0,
                tag=dataset_tag0
            )
            dataset_config_list.append(config)
        return DatasetConfigCollection(dataset_config_list=dataset_config_list, tag=collection_tag)

    def save_to_path(self, save_path: str, overwrite: bool=False):
        if file_exists(save_path) and not overwrite:
            logger.error(f'File already exists at: {save_path}')
            logger.error(f'Use overwrite=True to overwrite.')
            raise Exception
        extension = get_extension_from_path(save_path)
        if extension == 'json':
            json.dump(self.to_dict(), open(save_path, 'w'), indent=2, ensure_ascii=False)
        elif extension == 'yaml':
            yaml.dump(self.to_dict(), open(save_path, 'w'), allow_unicode=True)
        else:
            logger.error(f'Invalid file extension encountered: {extension}')
            logger.error(f"Valid file extensions: {['json', 'yaml']}")
            logger.error(f'Path specified: {save_path}')
            raise Exception

    @classmethod
    def load_from_path(cls, path: str) -> DatasetConfigCollection:
        check_file_exists(path)
        extension = get_extension_from_path(path)
        if extension == 'json':
            collection_dict = json.load(open(path, 'r'))
        elif extension == 'yaml':
            collection_dict = yaml.load(open(path, 'r'), Loader=yaml.FullLoader)
        else:
            logger.error(f'Invalid file extension encountered: {extension}')
            logger.error(f'Path specified: {path}')
            raise Exception
        return DatasetConfigCollection.from_dict(collection_dict)

    def filter_by_tag(self, tags: List[str]=None, new_collection_tag: str=None) -> DatasetConfigCollection:
        target_tags = [None] if tags is None else [tags] if type(tags) is not list else tags
        collection_tag = self.tag if new_collection_tag is None else new_collection_tag
        return DatasetConfigCollection(
            dataset_config_list=[config for config in self if config.tag in target_tags],
            tag=collection_tag
        )

class DatasetConfigCollectionHandler(BaseStructHandler['DatasetConfigCollectionHandler', 'DatasetConfigCollection']):
    def __init__(self, collection_list: List[DatasetConfigCollection]=None):
        collection_list0 = [collection for collection in collection_list if len(collection) > 0] if collection_list is not None else None
        super().__init__(obj_type=DatasetConfigCollection, obj_list=collection_list0)
        self.collection_list = self.obj_list

    def to_dict_list(self) -> List[dict]:
        return [collection.to_dict() for collection in self]

    @classmethod
    def from_dict_list(cls, collection_dict_list: List[dict]) -> DatasetConfigCollectionHandler:
        return DatasetConfigCollectionHandler(
            collection_list=[DatasetConfigCollection.from_dict(collection_dict) for collection_dict in collection_dict_list]
        )

    def save_to_path(self, save_path: str, overwrite: bool=False):
        if file_exists(save_path) and not overwrite:
            logger.error(f'File already exists at: {save_path}')
            logger.error(f'Use overwrite=True to overwrite.')
            raise Exception
        extension = get_extension_from_path(save_path)
        if extension == 'json':
            json.dump(self.to_dict_list(), open(save_path, 'w'), indent=2, ensure_ascii=False)
        elif extension == 'yaml':
            yaml.dump(self.to_dict_list(), open(save_path, 'w'), allow_unicode=True)
        else:
            logger.error(f'Invalid file extension encountered: {extension}')
            logger.error(f"Valid file extensions: {['json', 'yaml']}")
            logger.error(f'Path specified: {save_path}')
            raise Exception

    @classmethod
    def load_from_path(cls, path: str) -> DatasetConfigCollectionHandler:
        check_file_exists(path)
        extension = get_extension_from_path(path)
        if extension == 'json':
            collection_dict_list = json.load(open(path, 'r'))
        elif extension == 'yaml':
            collection_dict_list = yaml.load(open(path, 'r'), Loader=yaml.FullLoader)
        else:
            logger.error(f'Invalid file extension encountered: {extension}')
            logger.error(f'Path specified: {path}')
            raise Exception
        return DatasetConfigCollectionHandler.from_dict_list(collection_dict_list)

    def filter_by_collection_tag(self, tags: List[str]=None) -> DatasetConfigCollectionHandler:
        target_tags = [None] if tags is None else [tags] if type(tags) is not list else tags
        return DatasetConfigCollectionHandler([collection for collection in self if collection.tag in target_tags])

    def filter_by_dataset_tag(self, tags: List[str]=None) -> DatasetConfigCollectionHandler:
        target_tags = [None] if tags is None else [tags] if type(tags) is not list else tags
        return DatasetConfigCollectionHandler([collection.filter_by_tag(tags=target_tags) for collection in self])


from logger import logger
from xml.etree import ElementTree
from common_utils.path_utils import get_all_files_of_extension, get_rootname_from_path

class Source:
    def __init__(self, database: str):
        self.database = database

class Size:
    def __init__(self, width: int, height: int, depth: int):
        self.width = width
        self.height = height
        self.depth = depth

class BoundingBox:
    def __init__(self, xmin: int, ymin: int, xmax: int, ymax: int):
        self.xmin = xmin
        self.ymin = ymin
        self.xmax = xmax
        self.ymax = ymax

class Rectangle:
    def __init__(self, name: str, pose: str, truncated: int, difficult: int, bounding_box: BoundingBox):
        self.name = name
        self.pose = pose
        self.truncated = truncated
        self.difficult = difficult
        self.bounding_box = bounding_box

class RectangleHandler:
    def __init__(self):
        self.rectangles = []

    def add(self, rectangle: Rectangle):
        self.rectangles.append(rectangle)

class LabelImgAnnotationParser:
    def __init__(self, annotation_path: str):
        self.annotation_path = annotation_path

        self.folder = None
        self.filename = None
        self.path = None
        self.source = None
        self.size = None
        self.segmented = None
        self.rectangle_handler = None

    def load(self):
        self.load_data()
        logger.info(f"LabelImg Annotation Loaded: {self.annotation_path}")

    def load_data(self):
        tree = ElementTree.parse(source=self.annotation_path)
        root = tree.getroot()

        self.folder = root.find('folder').text
        self.filename = root.find('filename').text
        self.path = root.find('path').text
        source_tree = root.find('source')
        database = source_tree.find('database').text
        self.source = Source(database=database)
        size_tree = root.find('size')
        width = int(size_tree.find('width').text)
        height = int(size_tree.find('height').text)
        depth = int(size_tree.find('depth').text)
        self.size = Size(width=width, height=height, depth=depth)
        self.segmented = int(root.find('segmented').text)
        object_trees = root.findall('object')
        self.rectangle_handler = RectangleHandler()
        for object_tree in object_trees:
            name = object_tree.find('name').text
            pose = object_tree.find('pose').text
            truncated = int(object_tree.find('truncated').text)
            difficult = int(object_tree.find('difficult').text)
            bndbox_tree = object_tree.find('bndbox')
            xmin = int(bndbox_tree.find('xmin').text)
            ymin = int(bndbox_tree.find('ymin').text)
            xmax = int(bndbox_tree.find('xmax').text)
            ymax = int(bndbox_tree.find('ymax').text)
            bounding_box = BoundingBox(xmin=xmin, ymin=ymin, xmax=xmax, ymax=ymax)
            rectangle = Rectangle(
                name=name,
                pose=pose,
                truncated=truncated,
                difficult=difficult,
                bounding_box=bounding_box
            )
            self.rectangle_handler.add(rectangle)

class LabelImgAnnotation:
    def __init__(
        self, annotation_path: str, img_path: str
    ):
        self.annotation_path = annotation_path
        self.img_path = img_path

        self.folder = None
        self.filename = None
        self.path = None
        self.source = None
        self.size = None
        self.segmented = None
        self.rectangle_handler = None

    def load(self):
        parser = LabelImgAnnotationParser(self.annotation_path)
        parser.load()

        self.folder = parser.folder
        self.filename = parser.filename
        self.path = parser.path
        self.source = parser.source
        self.size = parser.size
        self.segmented = parser.segmented
        self.rectangle_handler = parser.rectangle_handler

    def unload(self):
        self.folder = None
        self.filename = None
        self.path = None
        self.source = None
        self.size = None
        self.segmented = None
        self.rectangle_handler = None

class LabelImgAnnotationHandler:
    def __init__(self):
        self.annotations = {}
        self.added_keys = []
        self.loaded_keys = []
        self.unloaded_keys = []

    def add(self, key, annotation_path: str, img_path: str):
        if key in self.annotations:
            logger.error(f"key={key} already exists in labelimg annotations dictionary")
            raise Exception
        self.annotations[key] = LabelImgAnnotation(annotation_path, img_path)
        self.added_keys.append(key)

    def delete(self, key):
        if key not in self.annotations:
            logger.error(f"key={key} doesn't exist in labelimg annotations dictionary")
            raise Exception
        del self.annotations[key]
        self.added_keys.remove(key)
        if key in self.loaded_keys:
            self.loaded_keys.remove(key)
        if key in self.unloaded_keys:
            self.unloaded_keys.remove(key)

    def load(self, key):
        if key not in self.annotations:
            logger.error(f"key={key} doesn't exist in labelimg annotations dictionary")
            raise Exception
        if key in self.loaded_keys:
            logger.error(f"key={key} has already been loaded in labelimg annotations dictionary")
            raise Exception
        self.annotations[key].load()
        self.loaded_keys.append(key)
        if key in self.unloaded_keys:
            self.unloaded_keys.remove(key)

    def unload(self, key):
        if key in self.annotations:
            logger.error(f"key={key} already exists in labelimg annotations dictionary")
            raise Exception
        if key not in self.loaded_keys:
            logger.error(f"key={key} hasn't been loaded in labelimg annotations dictionary yet")
            raise Exception
        self.annotations[key].unload()
        self.unloaded_keys.append(key)
        self.loaded_keys.remove(key)

    def load_list(self, key_list: list):
        for key in key_list:
            self.load(key)

    def unload_list(self, key_list: list):
        for key in key_list:
            self.unload(key)

    def load_remaining(self):
        key_list = [key for key in self.added_keys if key not in self.loaded_keys and key not in self.unloaded_keys]
        self.load_list(key_list)

    def unload_all(self):
        self.unload_list(self.loaded_keys)

    def load_batch(self, batch_size: int):
        remaining_keys = [key for key in self.added_keys if key not in self.loaded_keys and key not in self.unloaded_keys]
        if batch_size < len(remaining_keys):
            self.load_list(remaining_keys[:batch_size])
        else:
            self.load_list(remaining_keys)

    def finished_processing(self) -> bool:
        return len(self.added_keys) > 0 and self.unloaded_keys == self.added_keys

class LabelImgAnnotationHandlerTest:
    def __init__(self):
        self.handler = LabelImgAnnotationHandler()
        self.annotation_dir = "/home/clayton/workspace/prj/data_keep/youtube_promo/xml"
        self.img_dir = "/home/clayton/workspace/prj/data_keep/youtube_promo/img"

    def run(self):
        img_dirs = get_all_files_of_extension(dir_path=self.img_dir, extension='png')
        annotation_paths = get_all_files_of_extension(dir_path=self.annotation_dir, extension='xml')

        for i, annotation_path in zip(range(len(annotation_paths)), annotation_paths):
            rootname = get_rootname_from_path(path=annotation_path)
            img_path = f"{self.img_dir}/{rootname}.png"
            xml_path = f"{self.annotation_dir}/{rootname}.xml"
            self.handler.add(key=len(self.handler.annotations), annotation_path=xml_path, img_path=img_path)

        self.handler.load_remaining()



from common_utils.path_utils import get_all_files_of_extension, \
    get_rootname_from_path
from ..labelimg.annotation import LabelImgAnnotationHandler, LabelImgAnnotation
from ..labelme.annotation import LabelMeAnnotationHandler, LabelMeAnnotation, Shape, ShapeHandler
import labelme

class LabelImgLabelMeConverter:
    def __init__(
        self, labelimg_annotation_dir: str, labelme_annotation_dir: str, img_dir: str
    ):
        self.labelimg_annotation_dir = labelimg_annotation_dir
        self.labelme_annotation_dir = labelme_annotation_dir
        self.img_dir = img_dir

        # LabelImg Handler
        self.labelimg_annotation_handler = LabelImgAnnotationHandler()

    def load_annotation_paths(self):
        annotation_paths = get_all_files_of_extension(dir_path=self.labelimg_annotation_dir, extension='xml')
        for annotation_path in annotation_paths:
            rootname = get_rootname_from_path(path=annotation_path)
            img_path = f"{self.img_dir}/{rootname}.png"
            xml_path = f"{self.labelimg_annotation_dir}/{rootname}.xml"
            self.labelimg_annotation_handler.add(
                key=len(self.labelimg_annotation_handler.annotations),
                annotation_path=xml_path,
                img_path=img_path
            )

    def load_annotation_data(self):
        self.labelimg_annotation_handler.load_remaining()

    def load_labelimg(self):
        self.load_annotation_paths()
        self.load_annotation_data()

    def get_labelme_annotation_handler(self) -> LabelMeAnnotationHandler:
        labelme_annotation_handler = LabelMeAnnotationHandler()
        
        for labelimg_annotation in self.labelimg_annotation_handler.annotations.values():
            rootname = get_rootname_from_path(path=labelimg_annotation.annotation_path)
            labelme_annotation_path = f"{self.labelme_annotation_dir}/{rootname}.json"
            labelme_annotation = LabelMeAnnotation(
                annotation_path=labelme_annotation_path,
                img_dir=self.img_dir,
                bound_type='rect'
            )
            shape_handler = self.get_shape_handler(labelimg_annotation)
            shapes = self.get_shapes(shape_handler=shape_handler)

            labelme_annotation.version = labelme.__version__
            labelme_annotation.flags = {}
            labelme_annotation.shapes = shapes
            labelme_annotation.line_color = [0, 255, 0, 128]
            labelme_annotation.fill_color = [66, 255, 33, 128]
            labelme_annotation.img_path = labelimg_annotation.img_path
            labelme_annotation.img_height = labelimg_annotation.size.height
            labelme_annotation.img_width = labelimg_annotation.size.width
            labelme_annotation.shape_handler = shape_handler
        
            labelme_annotation_handler.annotations[
                len(labelme_annotation_handler.annotations)
            ] = labelme_annotation

        return labelme_annotation_handler

    def get_shape_handler(self, labelimg_annotation: LabelImgAnnotation) -> ShapeHandler:
        shape_handler = ShapeHandler()
        for rectangle in labelimg_annotation.rectangle_handler.rectangles:
            name = rectangle.name
            bounding_box = rectangle.bounding_box

            points = [
                [bounding_box.xmin, bounding_box.ymin],
                [bounding_box.xmax, bounding_box.ymax]
            ]
            shape = Shape(
                label=name,
                line_color=None,
                fill_color=None,
                points=points,
                shape_type='rectangle',
                flags=""
            )
            shape_handler.add(shape)
        return shape_handler

    def get_shapes(self, shape_handler: ShapeHandler):
        shapes = []
        for rectangle in shape_handler.rectangles:
            shape_dict = {}
            shape_dict['label'] = rectangle.label
            shape_dict['line_color'] = rectangle.line_color
            shape_dict['fill_color'] = rectangle.fill_color
            shape_dict['points'] = rectangle.points
            shape_dict['shape_type'] = rectangle.shape_type
            shape_dict['flags'] = {}
            shapes.append(shape_dict)
        return shapes
            


from __future__ import annotations
import json
from logger import logger
from common_utils.path_utils import rel_to_abs_path, get_dirpath_from_filepath

# TODO: These classes really need to be refactored.

class Shape:
    def __init__(self, label: str, group_id: int, points: list, shape_type: str, flags: str):
        if shape_type == 'point':
            self.hidden = True if label[0] == '&' else False
            self.label = label[1:] if self.hidden else label
        else:
            self.hidden = False
            self.label = label

        self.group_id = group_id
        self.points = points
        self.shape_type = shape_type
        self.flags = flags

    def __str__(self):
        return f"shape_type: {self.shape_type}, label: {self.label}, len(points): {len(self.points)}"

    def __repr__(self):
        return self.__str__()

    @classmethod
    def buffer(self, shape: Shape) -> Shape:
        return shape

    def copy(self) -> Shape:
        return Shape(
            label=self.label,
            group_id=self.group_id,
            points=self.points,
            shape_type=self.shape_type,
            flags=self.flags
        )

    def show_full(self):
        return f"shape_type: {self.shape_type}, label: {self.label}, points: {self.points}"

    def to_dict(self) -> dict:
        shape_dict = {}
        shape_dict['label'] = self.label if not self.hidden else f"&{self.label}"
        shape_dict['group_id'] = self.group_id
        shape_dict['points'] = self.points
        shape_dict['shape_type'] = self.shape_type
        shape_dict['flags'] = self.flags
        return shape_dict

class PolygonShape(Shape):
    def __init__(self, label: str, group_id: int, points: list, flags: str):
        super().__init__(
            label=label,
            group_id=group_id,
            points=points,
            shape_type='polygon',
            flags=flags
        )
        raise NotImplementedError

class RectangleShape(Shape):
    def __init__(self, label: str, group_id: int, points: list, flags: str):
        super().__init__(
            label=label,
            group_id=group_id,
            points=points,
            shape_type='rectangle',
            flags=flags
        )
        raise NotImplementedError

class CircleShape(Shape):
    def __init__(self, label: str, group_id: int, points: list, flags: str):
        super().__init__(
            label=label,
            group_id=group_id,
            points=points,
            shape_type='circle',
            flags=flags
        )
        raise NotImplementedError

class LineShape(Shape):
    def __init__(self, label: str, group_id: int, points: list, flags: str):
        super().__init__(
            label=label,
            group_id=group_id,
            points=points,
            shape_type='line',
            flags=flags
        )
        raise NotImplementedError

class PointShape(Shape):
    def __init__(self, label: str, group_id: int, points: list, flags: str):
        super().__init__(
            label=label,
            group_id=group_id,
            points=points,
            shape_type='point',
            flags=flags
        )
        raise NotImplementedError

class LinestripShape(Shape):
    def __init__(self, label: str, group_id: int, points: list, flags: str):
        super().__init__(
            label=label,
            group_id=group_id,
            points=points,
            shape_type='linestrip',
            flags=flags
        )
        raise NotImplementedError

class ShapeHandler:
    def __init__(self):
        self.polygons = []
        self.rectangles = []
        self.circles = []
        self.lines = []
        self.points = []
        self.linestrips = []

    def __str__(self):
        print_str = f"polygons:\n"
        for polygon in self.polygons:
            print_str += f"\t{polygon}\n"
        print_str += f"rectangles:\n"
        for rectangle in self.rectangles:
            print_str += f"\t{rectangle}\n"
        print_str += f"circles:\n"
        for circle in self.circles:
            print_str += f"\t{circle}\n"
        print_str += f"lines:\n"
        for line in self.lines:
            print_str += f"\t{line}\n"
        print_str += f"points:\n"
        for point in self.points:
            print_str += f"\t{point}\n"
        print_str += f"linestrips:\n"
        for linestrip in self.linestrips:
            print_str += f"\t{linestrip}\n"
        return print_str

    def __repr__(self):
        return self.__str__()

    def show_full(self):
        print_str = f"polygons:\n"
        for polygon in self.polygons:
            print_str += f"\t{polygon.show_full()}\n"
        print_str += f"rectangles:\n"
        for rectangle in self.rectangles:
            print_str += f"\t{rectangle.show_full()}\n"
        print_str += f"circles:\n"
        for circle in self.circles:
            print_str += f"\t{circle.show_full()}\n"
        print_str += f"lines:\n"
        for line in self.lines:
            print_str += f"\t{line.show_full()}\n"
        print_str += f"points:\n"
        for point in self.points:
            print_str += f"\t{point.show_full()}\n"
        print_str += f"linestrips:\n"
        for linestrip in self.linestrips:
            print_str += f"\t{linestrip.show_full()}\n"
        return print_str

    def add(self, shape: Shape):
        if shape.shape_type == 'polygon':
            self.polygons.append(shape)
        elif shape.shape_type == 'rectangle':
            self.rectangles.append(shape)
        elif shape.shape_type == 'circle':
            self.circles.append(shape)
        elif shape.shape_type == 'line':
            self.lines.append(shape)
        elif shape.shape_type == 'point':
            self.points.append(shape)
        elif shape.shape_type == 'linestrip':
            self.linestrips.append(shape)
        else:
            logger.error(f"Invalid shape_type: {shape.shape_type}")
            raise Exception

    def to_shape_list(self):
        shape_list = []
        for shape in self.polygons:
            shape_list.append(shape.to_dict())
        for shape in self.rectangles:
            shape_list.append(shape.to_dict())
        for shape in self.circles:
            shape_list.append(shape.to_dict())
        for shape in self.lines:
            shape_list.append(shape.to_dict())
        for shape in self.points:
            shape_list.append(shape.to_dict())
        for shape in self.linestrips:
            shape_list.append(shape.to_dict())
        return shape_list

class LabelMeAnnotationParser:
    def __init__(self, annotation_path: str):
        self.annotation_path = annotation_path

        self.version = None
        self.flags = None
        self.shapes = None
        self.img_path = None
        self.img_height = None
        self.img_width = None

        self.shape_handler = None

    def load(self):
        logger.purple(f"Loading: {self.annotation_path}")
        self.load_data()
        self.load_shapes()
        logger.info(f"LabelMe Annotation Loaded: {self.annotation_path}")

    def load_data(self):
        data = json.load(open(self.annotation_path, 'r'))
        self.version = data['version']
        self.flags = data['flags']
        self.shapes = data['shapes']
        line_color = data['lineColor'] if 'lineColor' in data else None
        fill_color = data['fillColor'] if 'fillColor' in data else None
        if line_color is not None or fill_color is not None:
            logger.warning(f"As of version 4.2.5, 'lineColor' and 'fillColor' keys in the global dictionary are decapritated.")
        img_path = data['imagePath']
        # get_dirpath_from_filepath(self.annotation_path)
        # self.img_path = rel_to_abs_path(f"{}/{img_path}")
        self.img_path = img_path
        self.img_height = data['imageHeight']
        self.img_width = data['imageWidth']

    def load_shapes(self):
        self.shape_handler = ShapeHandler()
        for shape in self.shapes:
            label = shape['label']
            line_color = shape['line_color'] if 'line_color' in shape else None # removed as of version 4.2.5
            fill_color = shape['fill_color'] if 'fill_color' in shape else None # removed as of version 4.2.5
            points = shape['points']
            group_id = shape['group_id'] if 'group_id' in shape else None # added as of version 4.2.5
            shape_type = shape['shape_type']
            flags = shape['flags'] if 'flags' in shape else None

            if line_color is not None or fill_color is not None:
                logger.warning(f"As of version 4.2.5, 'lineColor' and 'fillColor' keys in the shape dictionary are decapritated.")
            if flags is None:
                logger.warning(f"'flags' field not found in {self.annotation_path}")
                logger.warning(f"Ignoring flags field. Modify the json if you want to use flags.")
            shape_object = Shape( # TODO: Finish fixing
                label=label,
                group_id=group_id,
                points=points,
                shape_type=shape_type,
                flags=flags
            )
            self.shape_handler.add(shape_object)

    def shape_handler2shapes(self):
        self.shapes = self.shape_handler.to_shape_list()


class LabelMeAnnotation:
    def __init__(self, annotation_path: str, img_dir: str, bound_type: str):
        self.annotation_path = annotation_path
        self.img_dir = img_dir
        self.bound_type = bound_type

        self.version = None
        self.flags = None
        self.shapes = None
        self.img_path = None
        self.img_height = None
        self.img_width = None
        self.shape_handler = None

    def __str__(self):
        print_str = f"annotation_path: {self.annotation_path}"
        print_str += f"\nimg_dir: {self.img_dir}"
        print_str += f"\nbound_type: {self.bound_type}"
        print_str += f"\nversion: {self.version}"
        print_str += f"\nflags: {self.flags}"
        print_str += f"\nshapes: {self.shapes}"
        print_str += f"\nimg_path: {self.img_path}"
        print_str += f"\nimg_height: {self.img_height}"
        print_str += f"\nimg_width: {self.img_width}"
        if self.shape_handler is not None:
            print_str += f"\nshape_handler:\n{self.shape_handler.show_full()}"
        else:
            print_str += f"\nshape_handler:\n{self.shape_handler}"
        return print_str

    def __repr__(self):
        return self.__str__()

    def copy(self) -> LabelMeAnnotation:
        labelme_annotation = LabelMeAnnotation(
            annotation_path=self.annotation_path,
            img_dir=self.img_dir,
            bound_type=self.bound_type
        )
        labelme_annotation.version = self.version
        labelme_annotation.flags = self.flags
        labelme_annotation.shapes = self.shapes
        labelme_annotation.img_path = self.img_path
        labelme_annotation.img_height = self.img_height
        labelme_annotation.img_width = self.img_width
        labelme_annotation.shape_handler = self.shape_handler
        return labelme_annotation

    def load(self):
        parser = LabelMeAnnotationParser(self.annotation_path)
        parser.load()
        self.copy_from_parser(parser=parser)

    def copy_from_parser(self, parser: LabelMeAnnotationParser,
        annotation_path: str=None, img_dir: str=None, bound_type: str=None
    ):
        self.annotation_path = annotation_path if annotation_path is not None else self.annotation_path
        self.img_dir = img_dir if img_dir is not None else self.img_dir
        self.bound_type = bound_type if bound_type is not None else self.bound_type

        self.version = parser.version
        self.flags = parser.flags
        self.shapes = parser.shapes
        self.img_path = parser.img_path
        self.img_height = parser.img_height
        self.img_width = parser.img_width
        self.shape_handler = parser.shape_handler

    def unload(self):
        self.version = None
        self.flags = None
        self.shapes = None
        self.img_path = None
        self.img_height = None
        self.img_width = None
        self.shape_handler = None

class LabelMeAnnotationHandler:
    def __init__(self):
        self.annotations = {}
        self.added_keys = []
        self.loaded_keys = []
        self.unloaded_keys = []

    def add(self, key: str, annotation_path: str, img_dir: str, bound_type: str):
        if key in self.annotations:
            logger.error(f"key={key} already exists in labelme annotations dictionary")
            raise Exception
        self.annotations[key] = LabelMeAnnotation(annotation_path, img_dir, bound_type)
        self.added_keys.append(key)

    def delete(self, key: str):
        if key not in self.annotations:
            logger.error(f"key={key} doesn't exist in labelme annotations dictionary")
            raise Exception
        del self.annotations[key]
        self.added_keys.remove(key)
        if key in self.loaded_keys:
            self.loaded_keys.remove(key)
        if key in self.unloaded_keys:
            self.unloaded_keys.remove(key)

    def load(self, key: str):
        if key not in self.annotations:
            logger.error(f"key={key} doesn't exist in labelme annotations dictionary")
            raise Exception
        if key in self.loaded_keys:
            logger.error(f"key={key} has already been loaded in labelme annotations dictionary")
            raise Exception
        self.annotations[key].load()
        self.loaded_keys.append(key)
        if key in self.unloaded_keys:
            self.unloaded_keys.remove(key)

    def unload(self, key: str):
        if key in self.annotations:
            logger.error(f"key={key} already exists in labelme annotations dictionary")
            raise Exception
        if key not in self.loaded_keys:
            logger.error(f"key={key} hasn't been loaded in labelme annotations dictionary yet")
            raise Exception
        self.annotations[key].unload()
        self.unloaded_keys.append(key)
        self.loaded_keys.remove(key)

    def load_list(self, key_list: list):
        for key in key_list:
            self.load(key)

    def unload_list(self, key_list: list):
        for key in key_list:
            self.unload(key)

    def load_remaining(self):
        key_list = [key for key in self.added_keys if key not in self.loaded_keys and key not in self.unloaded_keys]
        self.load_list(key_list)

    def unload_all(self):
        self.unload_list(self.loaded_keys)

    def load_batch(self, batch_size: int):
        remaining_keys = [key for key in self.added_keys if key not in self.loaded_keys and key not in self.unloaded_keys]
        if batch_size < len(remaining_keys):
            self.load_list(remaining_keys[:batch_size])
        else:
            self.load_list(remaining_keys)

    def finished_processing(self) -> bool:
        return len(self.added_keys) > 0 and self.unloaded_keys == self.added_keys


import random
import numpy as np

from logger import logger

from common_utils.path_utils import get_next_dump_path
from common_utils.file_utils import dir_exists, delete_all_files_in_dir, make_dir_if_not_exists, make_dir, copy_file
from common_utils.path_utils import get_filename, get_dirpath_from_filepath, get_extension_from_path, get_extension_from_filename

from ..annotation import COCO_AnnotationFileParser
from ..structs import \
    COCO_License, COCO_Image, COCO_Annotation, COCO_Category
from ...util.utils.coco import COCO_Field_Buffer, COCO_Mapper_Handler
from ..writer import COCO_Writer

def get_possible_paths(path: str):
    path_parts = path.split('/')
    return ['/'.join(path_parts[i:]) for i in range(len(path_parts))]

def find_img_path(coco_url: str, src_root_dir: str) -> (bool, str):
    found = False
    img_path = None
    src_root_dirname = src_root_dir.split('/')[-1]
    for possible_path in get_possible_paths(path=coco_url):
        if possible_path.split('/')[0] == src_root_dirname:
            found = True
            img_path = f"{src_root_dir}/{'/'.join(possible_path.split('/')[1:])}"
            break
    return found, img_path

def print_id_update(label: str, added_new: bool, old_id: int, new_id: int):
    header = 'ADDED' if added_new else 'LINKED'
    logger.info(f"{label} {header}: {old_id} -> {new_id}")

class COCO_Splitter:
    def __init__(self, ann_path: list, src_root_dir: str, dest_dir: str, ratio: list=[2, 1, 0]):
        # Constructor Params
        self.ann_path = ann_path
        self.src_root_dir = src_root_dir
        self.dest_dir = dest_dir
        self.ratio = ratio

        # Parser Related
        parser = COCO_AnnotationFileParser(annotation_path=ann_path)
        parser.load(verbose=False)

        self.buffer = COCO_Field_Buffer.from_parser(
            parser=parser
        )
        self.train_buffer = COCO_Field_Buffer.from_scratch(
            description=f"Train COCO Dataset Split From {get_filename(path=ann_path)}",
            url=self.buffer.info.url,
            version=self.buffer.info.version
        )
        self.test_buffer = COCO_Field_Buffer.from_scratch(
            description=f"Test COCO Dataset Split From {get_filename(path=ann_path)}",
            url=self.buffer.info.url,
            version=self.buffer.info.version
        )
        self.val_buffer = COCO_Field_Buffer.from_scratch(
            description=f"Validation COCO Dataset Split From {get_filename(path=ann_path)}",
            url=self.buffer.info.url,
            version=self.buffer.info.version
        )

        self.buffer2train_mapper = COCO_Mapper_Handler()
        self.buffer2test_mapper = COCO_Mapper_Handler()
        self.buffer2val_mapper = COCO_Mapper_Handler()

        # Destination Directory Related
        self.train_dir = f"{self.dest_dir}/train"
        self.test_dir = f"{self.dest_dir}/test"
        self.val_dir = f"{self.dest_dir}/val"

        self.target_img_dir = "img"
        self.target_ann_dir = "ann"

    def setup_directories(self, verbose: bool=False):
        make_dir_if_not_exists(self.dest_dir)
        for target_dir in [self.train_dir, self.test_dir, self.val_dir]:
            if dir_exists(target_dir):
                delete_all_files_in_dir(dir_path=target_dir, ask_permission=False, verbose=verbose)
            else:
                make_dir(target_dir)
            img_dir = f"{target_dir}/{self.target_img_dir}"
            ann_dir = f"{target_dir}/{self.target_ann_dir}"
            make_dir(img_dir)
            make_dir(ann_dir)

    def split_image_list(self, image_list: list) -> (list, list, list):
        locations = np.cumsum([val*int(len(image_list)/sum(self.ratio)) for val in self.ratio]) - 1
        start_location = None
        end_location = 0
        count = 0
        sample_lists = []
        while count < len(locations):
            start_location = end_location
            end_location = locations[count]
            count += 1
            sample_lists.append(self.buffer.images.image_list[start_location:end_location])

        train_list, test_list, val_list = sample_lists
        return train_list, test_list, val_list

    def copy_image(self, coco_image: COCO_Image, dest_img_dir: str, new_img_filename: str=None, verbose: bool=False):
        coco_url = coco_image.coco_url
        found, src_img_path = find_img_path(coco_url=coco_url, src_root_dir=self.src_root_dir)
        if not found:
            logger.error(f"Couldn't find any permutation of coco_url={coco_url} under {self.src_root_dir}")
            raise Exception
        if new_img_filename is None:
            img_filename = get_filename(path=coco_url)
            if img_filename != coco_image.file_name:
                logger.error(f"coco_url filename doesn't match coco json file_name")
                logger.error(f"coco_url: {coco_url}")
                logger.error(f"file_name: {coco_image.file_name}")
                raise Exception
        else:
            img_filename = new_img_filename
        dest_img_path = f"{dest_img_dir}/{img_filename}"
        silent = not verbose
        copy_file(src_path=src_img_path, dest_path=dest_img_path, silent=silent)

    def update_images(
        self, target_buffer: COCO_Field_Buffer, target_mapper: COCO_Mapper_Handler,
        coco_image: COCO_Image, unique_key: str, new_img_filename: str=None, verbose: bool=False
    ):
        pending_coco_image = coco_image.copy()
        if new_img_filename is not None:
            coco_url = pending_coco_image.coco_url
            file_name = pending_coco_image.file_name
            coco_url_dirpath = get_dirpath_from_filepath(coco_url)
            coco_url_filename = get_filename(coco_url)
            if file_name != coco_url_filename:
                logger.error(f"file_name == {file_name} != coco_url_filename == {coco_url_filename}")
                raise Exception
            new_coco_url = f"{coco_url_dirpath}/{new_img_filename}"
            new_file_name = new_img_filename
            pending_coco_image.coco_url = new_coco_url
            pending_coco_image.file_name = new_file_name
            
        added_new, old_id, new_id = target_buffer.process_image(
            coco_image=pending_coco_image, id_mapper=target_mapper, unique_key=unique_key
        )
        if verbose:
            print_id_update(
                label=f'{unique_key} Image', added_new=added_new,
                old_id=old_id, new_id=new_id
            )

    def update_licenses(
        self, target_buffer: COCO_Field_Buffer, target_mapper: COCO_Mapper_Handler,
        coco_license: COCO_License, unique_key: str, verbose: bool=False
    ):
        added_new, old_id, new_id = target_buffer.process_license(
            coco_license=coco_license, id_mapper=target_mapper, unique_key=unique_key
        )
        if verbose:
            print_id_update(
                label=f'{unique_key} License', added_new=added_new,
                old_id=old_id, new_id=new_id
            )

    def update_annotations(
        self, target_buffer: COCO_Field_Buffer, target_mapper: COCO_Mapper_Handler,
        coco_annotation: COCO_Annotation, unique_key: str, verbose: bool=False
    ):
        added_new, old_id, new_id = target_buffer.process_annotation(
            coco_annotation=coco_annotation, id_mapper=target_mapper, unique_key=unique_key
        )
        if verbose:
            print_id_update(
                label=f'{unique_key} Annotation', added_new=added_new,
                old_id=old_id, new_id=new_id
            )

    def update_categories(
        self, target_buffer: COCO_Field_Buffer, target_mapper: COCO_Mapper_Handler,
        coco_category: COCO_Category, unique_key: str, verbose: bool=False
    ):
        added_new, old_id, new_id = target_buffer.process_category(
            coco_category=coco_category, id_mapper=target_mapper, unique_key=unique_key
        )
        if verbose:
            print_id_update(
                label=f'{unique_key} Category', added_new=added_new,
                old_id=old_id, new_id=new_id
            )

    def load_split(self, preserve_filenames: bool=False, verbose: bool=False):
        self.setup_directories(verbose=verbose)
        random.shuffle(self.buffer.images.image_list)
        train_image_list, test_image_list, val_image_list = self.split_image_list(image_list=self.buffer.images.image_list)

        for image_list, target_dir, target_buffer, target_mapper, unique_key in \
            zip(
                [train_image_list, test_image_list, val_image_list],
                [self.train_dir, self.test_dir, self.val_dir],
                [self.train_buffer, self.test_buffer, self.val_buffer],
                [self.buffer2train_mapper, self.buffer2test_mapper, self.buffer2val_mapper],
                ['train', 'test', 'val']
            ):
            dest_img_dir = f"{target_dir}/img"
            for coco_image in image_list:
                image_id = coco_image.id
                if not preserve_filenames:
                    img_extension = get_extension_from_filename(coco_image.file_name)
                    new_img_filename = get_filename(get_next_dump_path(dump_dir=dest_img_dir, file_extension=img_extension))
                    self.copy_image(coco_image=coco_image, dest_img_dir=dest_img_dir, new_img_filename=new_img_filename, verbose=verbose)
                else:
                    self.copy_image(coco_image=coco_image, dest_img_dir=dest_img_dir, verbose=verbose)
                coco_license = self.buffer.licenses.get_license_from_id(id=coco_image.license_id)
                self.update_licenses(
                    target_buffer=target_buffer, target_mapper=target_mapper,
                    coco_license=coco_license, unique_key=unique_key, verbose=verbose
                )
                if not preserve_filenames:
                    self.update_images(
                        target_buffer=target_buffer, target_mapper=target_mapper,
                        coco_image=coco_image, unique_key=unique_key, new_img_filename=new_img_filename, verbose=verbose
                    )
                else:
                    self.update_images(
                        target_buffer=target_buffer, target_mapper=target_mapper,
                        coco_image=coco_image, unique_key=unique_key, verbose=verbose
                    )
                coco_annotation_list = self.buffer.annotations.get_annotations_from_imgIds(imgIds=[image_id])
                for coco_annotation in coco_annotation_list:
                    coco_category = self.buffer.categories.get_category_from_id(id=coco_annotation.category_id)
                    self.update_categories(
                        target_buffer=target_buffer, target_mapper=target_mapper,
                        coco_category=coco_category, unique_key=unique_key, verbose=verbose
                    )
                    self.update_annotations(
                        target_buffer=target_buffer, target_mapper=target_mapper,
                        coco_annotation=coco_annotation, unique_key=unique_key, verbose=verbose
                    )
    
    def write_split(self, verbose: bool=False):
        for target_dir, target_buffer, unique_key in \
            zip(
                [self.train_dir, self.test_dir, self.val_dir],
                [self.train_buffer, self.test_buffer, self.val_buffer],
                ['train', 'test', 'val']
            ):
                if verbose:
                    logger.info(f"Writting {unique_key}.json from split...")
                ann_dir = f"{target_dir}/{self.target_ann_dir}"
                output_path = f"{ann_dir}/{unique_key}.json"
                writer = COCO_Writer.from_buffer(buffer=target_buffer, output_path=output_path)
                json_dict = writer.build_json_dict(verbose=verbose)
                writer.write_json_dict(json_dict=json_dict, verbose=verbose)
        if verbose:
            logger.good(f"Writing of split annotations is complete.")


from logger import logger
from ...labelme.annotation import Shape

def check_shape_type(shape: Shape, shape_type: str):
    if shape.shape_type != shape_type:
        logger.error(f"The given shape is not a {shape_type}.")
        logger.error(f"label: {shape.label}, shape_type: {shape.shape_type}")
        raise Exception


from logger import logger

from common_utils.check_utils import check_value_from_list
from common_utils.file_utils import delete_file
from ...coco.combiner import COCO_Annotations_Combiner
from ...coco.splitter import COCO_Splitter
from ..config import DatasetPathConfig

class DatasetCombineSplitWrapper:
    def __init__(
        self,
        dataset_path_config: str, src_root_dir: str, output_dir: str,
        ratio: list=[4, 0, 1], ann_format: str='coco'
    ):
        self.dataset_path_config = dataset_path_config
        self.src_root_dir = src_root_dir
        self.output_dir = output_dir
        self.ratio = ratio
        self.ann_format = ann_format
        self.combined_ann_output_path = f"/tmp/combined_ann_output.json"

    def run(self):
        # Load Dataset Paths From Config
        dataset_path_config = DatasetPathConfig.from_load(target=self.dataset_path_config)
        dataset_dir_list, img_dir_list, ann_path_list, ann_format_list = dataset_path_config.get_paths()
        check_value_from_list(item_list=ann_format_list, valid_value_list=[self.ann_format])

        # Combine Annotations
        worker = COCO_Annotations_Combiner(
            img_dir_list=img_dir_list,
            ann_path_list=ann_path_list
        )
        worker.load_combined(verbose=False, detailed_verbose=False)
        worker.write_combined(output_path=self.combined_ann_output_path, verbose=True)
        logger.purple(f"len(worker.licenses.license_list): {len(worker.buffer.licenses.license_list)}")
        logger.purple(f"len(worker.images.image_list): {len(worker.buffer.images.image_list)}")
        logger.purple(f"len(worker.annotations.annotation_list): {len(worker.buffer.annotations.annotation_list)}")
        logger.purple(f"len(worker.categories.category_list): {len(worker.buffer.categories.category_list)}")

        # Split Annotations In New Directory
        splitter = COCO_Splitter(
            ann_path=self.combined_ann_output_path,
            src_root_dir=self.src_root_dir,
            dest_dir=self.output_dir,
            ratio=self.ratio
        )
        delete_file(self.combined_ann_output_path)
        splitter.load_split(preserve_filenames=False, verbose=False)
        splitter.write_split(verbose=True)


from logger import logger
import json
from ..structs import COCO_Info, COCO_License_Handler, COCO_License, \
    COCO_Image_Handler, COCO_Image, COCO_Annotation_Handler, COCO_Annotation, \
    COCO_Category_Handler, COCO_Category
from ..camera import Camera

class COCO_AnnotationFileParser:
    def __init__(self, annotation_path: str):
        self.annotation_path = annotation_path

        self.info = None
        self.licenses = None
        self.images = None
        self.annotations = None
        self.categories = None

    def load(self, verbose: bool=False):
        self.load_data()
        if verbose: logger.info("COCO Annotation Loaded")

    def get_info(self, info: dict) -> COCO_Info:
        description = info['description']
        url = info['url']
        version = info['version']
        year = info['year']
        contributor = info['contributor']
        date_created = info['date_created']
        return COCO_Info(
            description=description,
            url=url,
            version=version,
            year=year,
            contributor=contributor,
            date_created=date_created
        )

    def get_licenses(self, licenses: list) -> COCO_License_Handler:
        coco_license_handler = COCO_License_Handler()
        for license_dict in licenses:
            url = license_dict['url']
            id = license_dict['id']
            name = license_dict['name']
            coco_license = COCO_License(
                url=url,
                id=id,
                name=name
            )
            coco_license_handler.add(coco_license)
        return coco_license_handler

    def get_images(self, images: list) -> COCO_Image_Handler:
        coco_image_handler = COCO_Image_Handler()
        for image_dict in images:
            license_id = image_dict['license']
            file_name = image_dict['file_name']
            coco_url = image_dict['coco_url'] if 'coco_url' in image_dict else None
            height = image_dict['height']
            width = image_dict['width']
            date_captured = image_dict['date_captured']
            flickr_url = image_dict['flickr_url'] if 'flickr_url' in image_dict else None
            id = image_dict['id']
            coco_image = COCO_Image(
                license_id=license_id,
                file_name=file_name,
                coco_url=coco_url,
                height=height,
                width=width,
                date_captured=date_captured,
                flickr_url=flickr_url,
                id=id
            )
            coco_image_handler.add(coco_image)
        return coco_image_handler

    def get_annotations(self, annotations: list):
        coco_annotation_handler = COCO_Annotation_Handler()
        for annotation_dict in annotations:
            segmentation = annotation_dict['segmentation']
            num_keypoints = annotation_dict['num_keypoints']
            area = annotation_dict['area']
            iscrowd = annotation_dict['iscrowd']
            keypoints = annotation_dict['keypoints']
            keypoints_3d = annotation_dict['keypoints_3d'] if 'keypoints_3d' in annotation_dict else None # Custom Field
            camera_params = annotation_dict['camera_params'] if 'camera_params' in annotation_dict else None # Custom Field
            image_id = annotation_dict['image_id']
            bbox = annotation_dict['bbox']
            category_id = annotation_dict['category_id']
            id = annotation_dict['id']
            coco_annotation = COCO_Annotation(
                segmentation=segmentation,
                num_keypoints=num_keypoints,
                area=area,
                iscrowd=iscrowd,
                keypoints=keypoints,
                image_id=image_id,
                bbox=bbox,
                category_id=category_id,
                id=id,
                keypoints_3d=keypoints_3d,
                camera=Camera.from_dict(intrinsic_param_dict=camera_params) if camera_params is not None else None
            )
            coco_annotation_handler.add(coco_annotation)
        return coco_annotation_handler

    def get_categories(self, categories: list) -> COCO_Category_Handler:
        coco_category_handler = COCO_Category_Handler()
        for category in categories:
            supercategory = category['supercategory']
            id = category['id']
            name = category['name']
            keypoints = category['keypoints']
            skeleton = category['skeleton']
            coco_category = COCO_Category(
                supercategory=supercategory,
                id=id,
                name=name,
                keypoints=keypoints,
                skeleton=skeleton
            )
            coco_category_handler.add(coco_category)
        return coco_category_handler

    def load_data(self):
        data = json.load(open(self.annotation_path, 'r'))
        info = data['info']
        self.info = self.get_info(info)
        licenses = data['licenses']
        self.licenses = self.get_licenses(licenses)
        images = data['images']
        self.images = self.get_images(images)
        annotations = data['annotations']
        self.annotations = self.get_annotations(annotations)
        categories = data['categories'] if 'categories' in data else None
        self.categories = self.get_categories(categories)


from __future__ import annotations
from typing import List
from logger import logger
from common_utils.common_types import Point, Keypoint, Rectangle
from common_utils.common_types.bbox import BBox
from common_utils.path_utils import get_extension_from_filename
from common_utils.check_utils import check_type

from ..camera import Camera

class COCO_Info:
    def __init__(
        self, description: str, url: str, version: str,
        year: str, contributor: str, date_created: str
    ):
        self.description = description
        self.url = url
        self.version = version
        self.year = year
        self.contributor = contributor
        self.date_created = date_created

    def __str__(self):
        print_str = ""
        print_str += f"description:\n\t{self.description}\n"
        print_str += f"url:\n\t{self.url}\n"
        print_str += f"version:\n\t{self.version}\n"
        print_str += f"year:\n\t{self.year}\n"
        print_str += f"contributor:\n\t{self.contributor}\n"
        print_str += f"date_created:\n\t{self.date_created}\n"
        return print_str

    def __repr__(self):
        return self.__str__()

    @classmethod
    def buffer(self, coco_info: COCO_Info) -> COCO_Info:
        return coco_info

    def copy(self) -> COCO_Info:
        return COCO_Info(
            description=self.description,
            url=self.url,
            version=self.version,
            year=self.year,
            contributor=self.contributor,
            date_created=self.date_created
        )

class COCO_License:
    def __init__(self, url: str, id: int, name: str):
        self.url = url
        self.id = id
        self.name = name

    def __str__(self):
        return f"url: {self.url}, id: {self.id}, name: {self.name}"

    def __repr__(self):
        return self.__str__()

    @classmethod
    def buffer(self, coco_license: COCO_License) -> COCO_License:
        return coco_license

    def copy(self) -> COCO_License:
        return COCO_License(
            url=self.url,
            id=self.id,
            name=self.name
        )

class COCO_License_Handler:
    def __init__(self):
        self.license_list = []

    def __str__(self):
        print_str = ""
        for coco_license in self.license_list:
            print_str += f"{coco_license}\n"

        return print_str

    def __repr__(self):
        return self.__str__()

    def __len__(self) -> int:
        return len(self.license_list)

    def __getitem__(self, idx: int) -> COCO_License:
        if len(self.license_list) == 0:
            logger.error(f"COCO_License_Handler is empty.")
            raise IndexError
        elif idx < 0 or idx >= len(self.license_list):
            logger.error(f"Index out of range: {idx}")
            raise IndexError
        else:
            return self.license_list[idx]

    def __setitem__(self, idx: int, value: COCO_License):
        check_type(value, valid_type_list=[COCO_License])
        self.license_list[idx] = value

    def __iter__(self):
        self.n = 0
        return self

    def __next__(self) -> COCO_License:
        if self.n < len(self.license_list):
            result = self.license_list[self.n]
            self.n += 1
            return result
        else:
            raise StopIteration

    def copy(self) -> COCO_License_Handler:
        result = COCO_License_Handler()
        result.license_list = self.license_list
        return result

    def add(self, coco_license: COCO_License):
        self.license_list.append(coco_license)

    def get_license_from_id(self, id: int) -> COCO_License:
        license_id_list = []
        for coco_license in self.license_list:
            if id == coco_license.id:
                return coco_license
            else:
                license_id_list.append(coco_license.id)
        license_id_list.sort()
        logger.error(f"Couldn't find coco_license with id={id}")
        logger.error(f"Possible ids: {license_id_list}")
        raise Exception

    def is_in_handler(self, coco_license: COCO_License, ignore_ids: bool=True, check_name_only: bool=False) -> bool:
        same_url = False
        same_id = False
        same_name = False

        found = False

        for existing_coco_license in self.license_list:
            same_url = True if existing_coco_license.url == coco_license.url else False
            same_id = True if existing_coco_license.id == coco_license.id else False
            same_name = True if existing_coco_license.name == coco_license.name else False
            if check_name_only:
                found = same_name
            else:
                if ignore_ids:
                    found = same_url and same_name
                else:
                    found = same_url and same_name and same_id
            if found:
                break
        return found

class COCO_Image:
    def __init__(
        self, license_id: int, file_name: str, coco_url: str,
        height: int, width: int, date_captured: str, flickr_url: str, id: int
    ):
        self.license_id = license_id
        self.file_name = file_name
        self.coco_url = coco_url
        self.height = height
        self.width = width
        self.date_captured = date_captured
        self.flickr_url = flickr_url
        self.id = id

    def __str__(self):
        print_str = "========================\n"
        print_str += f"license_id:\n\t{self.license_id}\n"
        print_str += f"file_name:\n\t{self.file_name}\n"
        print_str += f"coco_url:\n\t{self.coco_url}\n"
        print_str += f"height:\n\t{self.height}\n"
        print_str += f"width:\n\t{self.width}\n"
        print_str += f"date_captured:\n\t{self.date_captured}\n"
        print_str += f"flickr_url:\n\t{self.flickr_url}\n"
        print_str += f"id:\n\t{self.id}\n"
        return print_str

    def __repr__(self):
        return self.__str__()

    @classmethod
    def buffer(self, coco_image: COCO_Image) -> COCO_Image:
        return coco_image

    def copy(self) -> COCO_Image:
        return COCO_Image(
            license_id=self.license_id,
            file_name=self.file_name,
            coco_url=self.coco_url,
            height=self.height,
            width=self.width,
            date_captured=self.date_captured,
            flickr_url=self.flickr_url,
            id=self.id
        )

class COCO_Image_Handler:
    def __init__(self):
        self.image_list = []

    def __str__(self):
        print_str = ""
        for image in self.image_list:
            print_str += f"{image}\n"

        return print_str

    def __repr__(self):
        return self.__str__()

    def __len__(self) -> int:
        return len(self.image_list)

    def __getitem__(self, idx: int) -> COCO_Image:
        if len(self.image_list) == 0:
            logger.error(f"COCO_Image_Handler is empty.")
            raise IndexError
        elif idx < 0 or idx >= len(self.image_list):
            logger.error(f"Index out of range: {idx}")
            raise IndexError
        else:
            return self.image_list[idx]

    def __setitem__(self, idx: int, value: COCO_Image):
        check_type(value, valid_type_list=[COCO_Image])
        self.image_list[idx] = value

    def __iter__(self):
        self.n = 0
        return self

    def __next__(self) -> COCO_Image:
        if self.n < len(self.image_list):
            result = self.image_list[self.n]
            self.n += 1
            return result
        else:
            raise StopIteration

    def copy(self) -> COCO_Image_Handler:
        result = COCO_Image_Handler()
        result.image_list = self.image_list
        return result

    def add(self, coco_image: COCO_Image):
        self.image_list.append(coco_image)

    def get_image_from_id(self, id: int) -> COCO_Image:
        image_id_list = []
        for coco_image in self.image_list:
            if id == coco_image.id:
                return coco_image
            else:
                image_id_list.append(coco_image.id)
        image_id_list.sort()
        logger.error(f"Couldn't find coco_image with id={id}")
        logger.error(f"Possible ids: {image_id_list}")
        raise Exception

    def get_images_from_file_name(self, file_name: str) -> List[COCO_Image]:
        image_list = []
        for coco_image in self.image_list:
            if file_name == coco_image.file_name:
                image_list.append(coco_image)
        return image_list

    def get_images_from_coco_url(self, coco_url: str) -> List[COCO_Image]:
        image_list = []
        for coco_image in self.image_list:
            if coco_url == coco_image.coco_url:
                image_list.append(coco_image)
        return image_list

    def get_images_from_flickr_url(self, flickr_url: str) -> List[COCO_Image]:
        image_list = []
        for coco_image in self.image_list:
            if flickr_url == coco_image.flickr_url:
                image_list.append(coco_image)
        return image_list

    def get_extensions(self) -> List[str]:
        extension_list = []
        for coco_image in self.image_list:
            extension = get_extension_from_filename(coco_image.file_name)
            if extension not in extension_list:
                extension_list.append(extension)
        return extension_list

    def get_images_from_imgIds(self, imgIds: list) -> List[COCO_Image]:		
	        return [x for x in self.image_list if x.id in imgIds]

    def is_in_handler(self, coco_image: COCO_Image, ignore_ids: bool=True, check_file_name_only: bool=False) -> bool:
        same_license_id = False
        same_file_name = False
        same_coco_url = False
        same_height = False
        same_width = False
        same_date_captured = False
        same_flickr_url = False
        same_id = False

        found = False

        for existing_coco_image in self.image_list:
            same_file_name = True if existing_coco_image.file_name == coco_image.file_name else False
            same_coco_url = True if existing_coco_image.coco_url == coco_image.coco_url else False
            same_height = True if existing_coco_image.height == coco_image.height else False
            same_width = True if existing_coco_image.width == coco_image.width else False
            same_date_captured = True if existing_coco_image.date_captured == coco_image.date_captured else False
            same_flickr_url = True if existing_coco_image.flickr_url == coco_image.flickr_url else False
            same_license_id = True if existing_coco_image.license_id == coco_image.license_id else False
            same_id = True if existing_coco_image.id == coco_image.id else False
            if check_file_name_only:
                found = same_file_name
            else:
                if ignore_ids:
                    found = same_file_name and same_coco_url and same_height and same_width and \
                        same_date_captured and same_flickr_url
                else:
                    found = same_file_name and same_coco_url and same_height and same_width and \
                        same_date_captured and same_flickr_url and same_license_id and same_id
            if found:
                break
        return found


class COCO_Annotation:
    def __init__(
        self, segmentation: dict, num_keypoints: int, area: int, iscrowd: int,
        keypoints: list, image_id: int, bbox: list, category_id: int, id: int,
        keypoints_3d: list=None, camera: Camera=None
    ):
        self.segmentation = segmentation
        self.encoded_format = True if type(segmentation) is dict and 'size' in segmentation and 'counts' in segmentation else False
        self.num_keypoints = num_keypoints
        self.area = area
        self.iscrowd = iscrowd
        self.keypoints = keypoints
        self.image_id = image_id
        self.bbox = bbox
        self.category_id = category_id
        self.id = id

        self.keypoints_3d = keypoints_3d
        self.camera = camera

        self.seg_point_lists = self.get_seg_point_lists()
        self.keypoint_list = self.get_keypoint_list()
        self.bounding_box = self.get_bounding_box()

    def __str__(self):
        print_str = "========================\n"
        print_str += f"encoded_format:\n\t{self.encoded_format}\n"
        print_str += f"segmentation:\n\t{self.segmentation}\n"
        print_str += f"num_keypoints:\n\t{self.num_keypoints}\n"
        print_str += f"area:\n\t{self.area}\n"
        print_str += f"iscrowd:\n\t{self.iscrowd}\n"
        print_str += f"keypoints:\n\t{self.keypoints}\n"
        print_str += f"image_id:\n\t{self.image_id}\n"
        print_str += f"bbox:\n\t{self.bbox}\n"
        print_str += f"category_id:\n\t{self.category_id}\n"
        print_str += f"id:\n\t{self.id}\n"
        return print_str

    def __repr__(self):
        return self.__str__()

    @classmethod
    def buffer(self, coco_annotation: COCO_Annotation) -> COCO_Annotation:
        return coco_annotation

    def copy(self) -> COCO_Annotation:
        return COCO_Annotation(
            segmentation=self.segmentation,
            num_keypoints=self.num_keypoints,
            area=self.area,
            iscrowd=self.iscrowd,
            keypoints=self.keypoints,
            image_id=self.image_id,
            bbox=self.bbox,
            category_id=self.category_id,
            id=self.id,
            keypoints_3d=self.keypoints_3d,
            camera=self.camera
        )

    def _chunk2pointlist(self, chunk: list) -> list:
        if len(chunk) % 2 != 0:
            logger.warning(f"Chunk is not evenly divisible by 2. len(chunk)={len(chunk)}")
            return []
        point_list = []
        for i in range(int(len(chunk)/2)):
            coord = chunk[2*i:2*(i+1)]
            point = Point(coord[0], coord[1])
            point_list.append(point)
        return point_list

    def get_seg_point_lists(self) -> list:
        if self.encoded_format:
            # TODO: Figure out how to convert the encoded format to the standard format.
            #logger.warning(f"Not implemented yet for encoded segmentation format.")
            return []
        else:
            point_lists = []
            for value_list in self.segmentation:
                pointlist = self._chunk2pointlist(value_list)
                point_lists.append(pointlist)
            return point_lists

    def get_keypoint_list(self) -> list:
        if self.keypoints is None:
            if self.num_keypoints > 0:
                logger.warning(f"No keypoints provided to COCO_Annotation constructor. Expecting {self.num_keypoints}.")
            return []
        if len(self.keypoints) % 3 != 0:
            logger.warning(f"self.keypoints is not evenly divisible by 3. len(self.keypoints)={len(self.keypoints)}")
            return []
        keypoint_list = []
        for i in range(int(len(self.keypoints)/3)):
            coord = self.keypoints[3*i:3*(i+1)]
            keypoint = Keypoint(coord[0], coord[1], coord[2])
            keypoint_list.append(keypoint)
        return keypoint_list

    def get_keypoint_3d_list(self) -> list:
        if self.keypoints_3d is None:
            return []
        if len(self.keypoints_3d) % 4 != 0:
            logger.warning(f"self.keypoints_3d is not evenly divisible by 4. len(self.keypoints_3d)={len(self.keypoints_3d)}")
            return []
        keypoint_3d_list = []
        for i in range(int(len(self.keypoints_3d)/4)):
            coord = self.keypoints_3d[4*i:4*(i+1)]
            keypoint_3d_list.append(coord)
        return keypoint_3d_list

    def get_bounding_box(self) -> BBox:
        return BBox.from_list([self.bbox[0], self.bbox[1], self.bbox[0]+self.bbox[2], self.bbox[1]+self.bbox[3]])

    def get_rect(self) -> Rectangle:
        return Rectangle(self.bbox[0], self.bbox[1], self.bbox[0]+self.bbox[2], self.bbox[1]+self.bbox[3])

class COCO_Annotation_Handler:
    def __init__(self):
        self.annotation_list = []

    def __str__(self):
        print_str = ""
        for annotation in self.annotation_list:
            print_str += f"{annotation}\n"

        return print_str

    def __repr__(self):
        return self.__str__()

    def __len__(self) -> int:
        return len(self.annotation_list)

    def __getitem__(self, idx: int) -> COCO_Annotation:
        if len(self.annotation_list) == 0:
            logger.error(f"COCO_Annotation_Handler is empty.")
            raise IndexError
        elif idx < 0 or idx >= len(self.annotation_list):
            logger.error(f"Index out of range: {idx}")
            raise IndexError
        else:
            return self.annotation_list[idx]

    def __setitem__(self, idx: int, value: COCO_Annotation):
        check_type(value, valid_type_list=[COCO_Annotation])
        self.annotation_list[idx] = value

    def __iter__(self):
        self.n = 0
        return self

    def __next__(self) -> COCO_Annotation:
        if self.n < len(self.annotation_list):
            result = self.annotation_list[self.n]
            self.n += 1
            return result
        else:
            raise StopIteration

    def copy(self) -> COCO_Annotation_Handler:
        result = COCO_Annotation_Handler()
        result.annotation_list = self.annotation_list
        return result

    def add(self, coco_annotation: COCO_Annotation):
        self.annotation_list.append(coco_annotation)

    def get_annotation_from_id(self, id: int) -> COCO_Annotation:
        annotation_id_list = []
        for coco_annotation in self.annotation_list:
            if id == coco_annotation.id:
                return coco_annotation
            else:
                annotation_id_list.append(coco_annotation.id)
        annotation_id_list.sort()
        logger.error(f"Couldn't find coco_annotation with id={id}")
        logger.error(f"Possible ids: {annotation_id_list}")
        raise Exception

    def get_annotations_from_annIds(self, annIds: list) -> List[COCO_Annotation]:		
        return [x for x in self.annotation_list if x.id in annIds]		
        
    def get_annotations_from_imgIds(self, imgIds: list) -> List[COCO_Annotation]:		
        return [x for x in self.annotation_list if x.image_id in imgIds]

class COCO_Category:
    def __init__(
        self, supercategory: str, id: int, name: str, keypoints: list, skeleton: list
    ):
        self.supercategory = supercategory
        self.id = id
        self.name = name
        self.keypoints = keypoints
        self.skeleton = skeleton

        self.label_skeleton = self.get_label_skeleton()

    def __str__(self):
        print_str = "========================\n"
        print_str += f"supercategory:\n\t{self.supercategory}\n"
        print_str += f"id:\n\t{self.id}\n"
        print_str += f"name:\n\t{self.name}\n"
        print_str += f"keypoints:\n\t{self.keypoints}\n"
        print_str += f"skeleton:\n\t{self.skeleton}\n"
        return print_str

    def __repr__(self):
        return self.__str__()

    @classmethod
    def buffer(self, coco_category: COCO_Category) -> COCO_Category:
        return coco_category

    def copy(self) -> COCO_Category:
        return COCO_Category(
            supercategory=self.supercategory,
            id=self.id,
            name=self.name,
            keypoints=self.keypoints,
            skeleton=self.skeleton
        )

    def get_label_skeleton(self) -> list:
        str_skeleton = []
        for int_bone in self.skeleton:
            bone_start = self.keypoints[int_bone[0]-1] 
            bone_end = self.keypoints[int_bone[1]-1]
            str_bone = [bone_start, bone_end]
            str_skeleton.append(str_bone)
        return str_skeleton

class COCO_Category_Handler:
    def __init__(self):
        self.category_list = []

    def __str__(self):
        print_str = ""
        for category in self.category_list:
            print_str += f"{category}\n"

        return print_str

    def __repr__(self):
        return self.__str__()

    def __len__(self) -> int:
        return len(self.category_list)

    def __getitem__(self, idx: int) -> COCO_Category:
        if len(self.category_list) == 0:
            logger.error(f"COCO_Category_Handler is empty.")
            raise IndexError
        elif idx < 0 or idx >= len(self.category_list):
            logger.error(f"Index out of range: {idx}")
            raise IndexError
        else:
            return self.category_list[idx]

    def __setitem__(self, idx: int, value: COCO_Category):
        check_type(value, valid_type_list=[COCO_Category])
        self.category_list[idx] = value

    def __iter__(self):
        self.n = 0
        return self

    def __next__(self) -> COCO_Category:
        if self.n < len(self.category_list):
            result = self.category_list[self.n]
            self.n += 1
            return result
        else:
            raise StopIteration

    def copy(self) -> COCO_Category_Handler:
        result = COCO_Category_Handler()
        result.category_list = self.category_list
        return result

    def add(self, coco_category: COCO_Category):
        self.category_list.append(coco_category)

    def get_categories_from_name(self, name: str) -> List[COCO_Category]:		
        return [x for x in self.category_list if x.name == name]

    def get_unique_category_from_name(self, name: str) -> COCO_Category:
        found_categories = self.get_categories_from_name(name)
        if len(found_categories) == 0:
            logger.error(f"Couldn't find any categories by the name: {name}")
            raise Exception
        elif len(found_categories) > 1:
            logger.error(f"Found {len(found_categories)} categories with the name {name}")
            logger.error(f"Found Categories:")
            for category in found_categories:
                logger.error(category)
            raise Exception
        return found_categories[0]

    def get_category_from_id(self, id: int) -> COCO_Category:
        category_id_list = []
        for coco_category in self.category_list:
            if id == coco_category.id:
                return coco_category
            else:
                category_id_list.append(coco_category.id)
        category_id_list.sort()
        logger.error(f"Couldn't find coco_category with id={id}")
        logger.error(f"Possible ids: {category_id_list}")
        raise Exception

    def get_skeleton_from_name(self, name: str) -> (list, list):
        unique_category = self.get_unique_category_from_name(name)
        skeleton = unique_category.skeleton
        label_skeleton = unique_category.get_label_skeleton()
        return skeleton, label_skeleton

    def is_in_handler(self, coco_category: COCO_Category, ignore_ids: bool=True, check_name_only: bool=True) -> bool:
        same_supercategory = False
        same_id = False
        same_name = False
        same_keypoints = False
        same_skeleton = False

        found = False

        for existing_coco_category in self.category_list:
            same_supercategory = True if existing_coco_category.supercategory == coco_category.supercategory else False
            same_id = True if existing_coco_category.id == coco_category.id else False
            same_name = True if existing_coco_category.name == coco_category.name else False
            same_keypoints = True if existing_coco_category.keypoints == coco_category.keypoints else False
            same_skeleton = True if existing_coco_category.skeleton == coco_category.skeleton else False
            if check_name_only:
                found = same_name
            else:
                if ignore_ids:
                    found = same_supercategory and same_name and same_keypoints and same_skeleton
                else:
                    found = same_supercategory and same_name and same_keypoints and same_skeleton and same_id
            if found:
                break
        return found


from logger import logger
from common_utils.path_utils import get_all_files_of_extension, get_rootname_from_path, \
    get_filename, get_dirpath_from_filepath
from common_utils.file_utils import file_exists
from common_utils.time_utils import get_present_year, get_present_time_Ymd, \
    get_ctime
from common_utils.user_utils import get_username
from ..labelme.annotation import LabelMeAnnotationHandler
from ..coco.structs import COCO_Info, COCO_License_Handler, COCO_License, \
    COCO_Image_Handler, COCO_Image, COCO_Annotation_Handler, COCO_Annotation, \
    COCO_Category_Handler, COCO_Category
from ..util.utils import polygon2segmentation, polygon2bbox, point_inside_polygon, \
    labeled_points2keypoints, rectangle2bbox, point_inside_rectangle

class LabelMeCOCOConverter:
    def __init__(
        self, labelme_img_dir_list: list, labelme_annotation_dir_list: list, labelme_index_range_list: list=None,
        bound_type_list: str=None, category_dict: dict=None
    ):
        # Image Directory List (for getting dates modified)
        self.labelme_img_dir_list = labelme_img_dir_list

        # Category Dictionary (for handling keypoints)
        self.category_dict = category_dict

        # Bound type used for grouping keypoints
        self.bound_type_list = bound_type_list if bound_type_list is not None else ['poly'] * len(self.labelme_img_dir_list)

        # Annotation Paths List
        self.labelme_annotation_dir_list = labelme_annotation_dir_list
        self.annotation_pathlist_list = []
        for labelme_annotation_dir in labelme_annotation_dir_list:
            annotation_pathlist = get_all_files_of_extension(
                dir_path=labelme_annotation_dir,
                extension="json"
            )
            self.annotation_pathlist_list.append(annotation_pathlist)

        # Index ranges to be loaded
        self.labelme_index_range_list = labelme_index_range_list if labelme_index_range_list is not None else ['all'] * len(self.labelme_img_dir_list)

        # LabelMe Handler
        self.labelme_annotation_handler = LabelMeAnnotationHandler()

        # COCO Data
        self.info = None
        self.licenses = None
        self.images = None
        self.annotations = None
        self.categories = None

        # For keeping the ids unique
        self.img_path2id_dict = {}

    def load_annotation_paths(self):
        for i, annotation_pathlist in zip(range(len(self.annotation_pathlist_list)), self.annotation_pathlist_list):
            index_range = self.labelme_index_range_list[i]
            if index_range == 'none':
                continue
            elif index_range == 'all':
                for annotation_path in annotation_pathlist:
                    self.labelme_annotation_handler.add(
                        key=len(self.labelme_annotation_handler.annotations), annotation_path=annotation_path, img_dir=self.labelme_img_dir_list[i],
                        bound_type=self.bound_type_list[i]
                    )
            else:
                for annotation_path in annotation_pathlist:
                    annotation_index = int(get_rootname_from_path(annotation_path))
                    for start_index, end_index in index_range:
                        if annotation_index >= start_index and annotation_index <= end_index:
                            self.labelme_annotation_handler.add(
                                key=len(self.labelme_annotation_handler.annotations), annotation_path=annotation_path, img_dir=self.labelme_img_dir_list[i],
                                bound_type=self.bound_type_list[i]
                            )
                            break

    def load_annotation_data(self):
        self.labelme_annotation_handler.load_remaining()

    def load_labelme(self):
        self.load_annotation_paths()
        self.load_annotation_data()

    def determine_coco_fields(self):
        self.info = self.get_info()
        self.licenses = self.get_licenses()
        self.images = self.get_images()
        self.categories = self.get_categories()
        self.annotations = self.get_annotations()

    def get_info(self) -> COCO_Info:
        return COCO_Info(
            description="COCO Dataset generated from labelme tool",
            url="https://github.com/wkentaro/labelme",
            version="1.0",
            year=get_present_year(),
            contributor=get_username(),
            date_created=get_present_time_Ymd()
        )

    def get_licenses(self) -> COCO_License_Handler:
        coco_license_handler = COCO_License_Handler()
        labelme_license = COCO_License(
            url="https://github.com/wkentaro/labelme/blob/master/LICENSE",
            id=1,
            name="GNU General Public License"
        )
        coco_license_handler.add(labelme_license)
        return coco_license_handler

    def get_images(self) -> COCO_Image_Handler:
        coco_image_handler = COCO_Image_Handler()
        for annotation in self.labelme_annotation_handler.annotations.values():
            height = annotation.img_height
            filename = get_filename(annotation.img_path)
            logger.yellow(f"annotation.img_dir: {annotation.img_dir}")
            logger.purple(f"annotation.img_path: {annotation.img_path}")
            img_path = f"{annotation.img_dir}/{filename}"
            if file_exists(img_path):
                date_captured = get_ctime(img_path)
            else:
                assumed_img_dir = get_dirpath_from_filepath(annotation.img_path)
                logger.warning(f"Image directory not specified.")
                logger.warning(f"Assuming img_dir = {assumed_img_dir}")
                date_captured = get_ctime(annotation.img_path)

            id = len(self.img_path2id_dict)
            self.img_path2id_dict[img_path] = id
            coco_image = COCO_Image(
                license_id=1,
                file_name=filename,
                coco_url=img_path,
                height=annotation.img_height,
                width=annotation.img_width,
                date_captured=date_captured,
                flickr_url=None,
                id=id
            )
            coco_image_handler.add(coco_image)
        return coco_image_handler

    def get_annotations(self) -> COCO_Annotation_Handler:
        """
        Note: There is no indication of whether two separate segmentations
        correspond to the same object in labelme, so for now I will just
        make it so that all segmentations are independent.
        In other words, each segmentation list will only have at most one
        element.
        TODO: Figure out how to resolve the above mentioned issue.
        """
        coco_annotation_handler = COCO_Annotation_Handler()
        for annotation in self.labelme_annotation_handler.annotations.values():
            img_filename = get_filename(annotation.img_path)
            img_path = f"{annotation.img_dir}/{img_filename}"
            img_id = self.img_path2id_dict[img_path]
            logger.info(f"Retrieving Annotation: {img_id}")
            # here
            # circles = annotation.shape_handler.circles
            # lines = annotation.shape_handler.lines
            # linestrips = annotation.shape_handler.linestrips

            seg_dict = {}
            if annotation.bound_type == 'poly':
                polygons = annotation.shape_handler.polygons
                if len(polygons) > 0:
                    for polygon in polygons:
                        segmentation_list = []
                        segmentation = polygon2segmentation(polygon)
                        segmentation_list.append(segmentation)
                        seg_bbox = polygon2bbox(polygon)
                        seg_dict[len(seg_dict)] = {
                            'segmentation_list': segmentation_list,
                            'seg_bbox_area': seg_bbox.area,
                            'seg_bbox': seg_bbox,
                            'shape_obj': polygon,
                            'labeled_points': [],
                            'label': polygon.label
                        }
            elif annotation.bound_type == 'rect':
                rectangles = annotation.shape_handler.rectangles
                if len(rectangles) > 0:
                    for rectangle in rectangles:
                        bbox = rectangle2bbox(rectangle=rectangle)
                        seg_dict[len(seg_dict)] = {
                            'segmentation_list': [],
                            'seg_bbox_area': bbox.area,
                            'seg_bbox': bbox,
                            'shape_obj': rectangle,
                            'labeled_points': [],
                            'label': rectangle.label
                        }

            points = annotation.shape_handler.points
            if len(points) > 0:
                for point in points:
                    for seg_item in seg_dict.values():
                        if annotation.bound_type == 'poly':
                            if point_inside_polygon(
                                point=point,
                                polygon=seg_item['shape_obj']
                            ):
                                seg_item['labeled_points'].append(point)
                        elif annotation.bound_type == 'rect':
                            if point_inside_rectangle(
                                point=point,
                                rectangle=seg_item['shape_obj']
                            ):
                                seg_item['labeled_points'].append(point)
                        else:
                            logger.error(f"Invalid bound_type: {annotation.bound_type}")
                            raise Exception

            for seg_item in seg_dict.values():
                num_keypoints = len(seg_item['labeled_points'])
                keypoint_labels = None
                category_found = False
                category_id = None
                for category in self.categories.category_list:
                    if category.name == seg_item['label']:
                        keypoint_labels = category.keypoints
                        category_found = True
                        category_id = category.id
                        break
                if not category_found:
                    logger.warning(f"Warning: Category label {seg_item['label']} not found.")
                    logger.warning(f"Ignoring {len(seg_item['labeled_points'])} labeled points.")

                keypoints = labeled_points2keypoints(
                    keypoint_labels=keypoint_labels,
                    labeled_points=seg_item['labeled_points'],
                    img_path=img_path,
                    annotation_path=annotation.annotation_path
                ) if keypoint_labels is not None else None
                if num_keypoints > 0 and keypoint_labels is None:
                    logger.warning(f"Warning: No keypoint_labels specified.")
                    logger.warning(f"Cannot generate keypoints without keypoint_labels.")
                    logger.warning(f"{num_keypoints} labeled_points ignored.")
                
                image_id = img_id
                seg_bbox = seg_item['seg_bbox']
                coco_annotation = COCO_Annotation(
                    segmentation=seg_item['segmentation_list'],
                    num_keypoints=num_keypoints,
                    area=seg_item['seg_bbox_area'],
                    iscrowd=0,
                    keypoints=keypoints,
                    image_id=image_id,
                    bbox=[seg_bbox.xmin, seg_bbox.ymin, seg_bbox.width, seg_bbox.height],
                    category_id=category_id,
                    id=len(coco_annotation_handler.annotation_list)
                )
                coco_annotation_handler.add(coco_annotation)
        return coco_annotation_handler

    def get_categories(self) -> COCO_Category_Handler:
        coco_category_handler = COCO_Category_Handler()        
        if self.category_dict is None:
            logger.warning("Warning: No category_dict provided.")
            logger.warning("All keypoints will be ignored.")
            return coco_category_handler
        for key, data in self.category_dict.items():
            coco_category = COCO_Category(
                supercategory=key,
                id=len(coco_category_handler.category_list),
                name=key,
                keypoints=data['keypoints'],
                skeleton=data['skeleton']
            )
            coco_category_handler.add(coco_category)
        return coco_category_handler

    def test(self):
        self.load_labelme()
        self.determine_coco_fields()
        logger.cyan("Info")
        logger.blue(self.info)
        logger.cyan("Licenses")
        logger.blue(self.licenses)
        logger.cyan("Images")
        logger.blue(self.images)
        logger.cyan("Annotations")
        logger.blue(self.annotations)
        logger.cyan("Categories")
        logger.blue(self.categories)



from __future__ import annotations

from typing import List
import json
import operator
import random

from logger import logger
from common_utils.check_utils import check_type, check_type_from_list, \
    check_file_exists, check_value_from_list
from common_utils.path_utils import get_extension_from_filename
from common_utils.file_utils import file_exists

from .objects import COCO_License, COCO_Image, COCO_Annotation, COCO_Category
from ....base import BaseStructHandler

class COCO_License_Handler(BaseStructHandler['COCO_License_Handler', 'COCO_License']):
    def __init__(self, license_list: List[COCO_License]=None):
        super().__init__(obj_type=COCO_License, obj_list=license_list)
        self.license_list = self.obj_list

    @classmethod
    def from_dict_list(cls, dict_list: List[dict]) -> COCO_License_Handler:
        return COCO_License_Handler(
            license_list=[COCO_License.from_dict(license_dict) for license_dict in dict_list]
        )

    @classmethod
    def load_from_path(cls, json_path: str) -> COCO_License_Handler:
        check_file_exists(json_path)
        json_data = json.load(open(json_path, 'r'))
        return COCO_License_Handler.from_dict_list(json_data)

    def remove(self, id_list: List[int], verbose: bool=False):
        # TODO: Create a base class that inherits from BaseStruct that requires an id class parameter
        # This method could be added to the base handler of the resulting object class.
        if len(id_list) > 0:
            idx_list = list(range(len(self)))
            idx_list.reverse()
            for idx in idx_list:
                if self[idx].id in id_list:
                    del self[idx]
                    if verbose:
                        logger.info(f'Deleted License Id: {idx}')

    def remove_if_no_imgs(self, img_handler: COCO_Image_Handler, id_list: List[int]=None, verbose: bool=False):
        rm_license_id_list = []
        if id_list is not None:
            for license_id in id_list:
                imgs = img_handler.get_images_from_licenseIds([license_id])
                if len(imgs) == 0:
                    rm_license_id_list.append(license_id)
        else:
            for coco_license in self:
                imgs = img_handler.get_images_from_licenseIds([coco_license.id])
                if len(imgs) == 0:
                    rm_license_id_list.append(license_id)
        self.remove(rm_license_id_list, verbose=verbose)

class COCO_Image_Handler(BaseStructHandler['COCO_Image_Handler', 'COCO_Image']):
    def __init__(self, image_list: List[COCO_Image]=None):
        super().__init__(obj_type=COCO_Image, obj_list=image_list)
        self.image_list = self.obj_list

    def get_images_from_file_name(self, file_name: str) -> List[COCO_Image]:
        return [coco_image for coco_image in self if file_name == coco_image.file_name]

    def get_images_from_coco_url(self, coco_url: str) -> List[COCO_Image]:
        return [coco_image for coco_image in self if coco_url == coco_image.coco_url]

    def get_images_from_flickr_url(self, flickr_url: str) -> List[COCO_Image]:
        return [coco_image for coco_image in self if flickr_url == coco_image.flickr_url]

    def get_extensions(self) -> List[str]:
        extension_list = []
        for coco_image in self:
            extension = get_extension_from_filename(coco_image.file_name)
            if extension not in extension_list:
                extension_list.append(extension)
        return extension_list

    def get_images_from_imgIds(self, imgIds: list) -> List[COCO_Image]:		
	        return [x for x in self if x.id in imgIds]

    def get_images_from_licenseIds(self, licenseIds: List[int]) -> List[COCO_Image]:
        return [x for x in self if x.license_id in licenseIds]

    @classmethod
    def from_dict_list(cls, dict_list: List[dict]) -> COCO_Image_Handler:
        return COCO_Image_Handler(
            image_list=[COCO_Image.from_dict(image_dict) for image_dict in dict_list]
        )

    @classmethod
    def load_from_path(cls, json_path: str) -> COCO_Image_Handler:
        check_file_exists(json_path)
        json_data = json.load(open(json_path, 'r'))
        return COCO_Image_Handler.from_dict_list(json_data)

    def remove(self, id_list: List[int], verbose: bool=False):
        # TODO: Create a base class that inherits from BaseStruct that requires an id class parameter
        # This method could be added to the base handler of the resulting object class.
        if len(id_list) > 0:
            idx_list = list(range(len(self)))
            idx_list.reverse()
            for idx in idx_list:
                if self[idx].id in id_list:
                    del self[idx]
                    if verbose:
                        logger.info(f'Deleted Image Id: {idx}')
    
    def remove_if_no_anns(self, ann_handler: COCO_Annotation_Handler, license_handler: COCO_License_Handler=None, id_list: List[int]=None, verbose: bool=False):
        """Removes all of the COCO_Image objects in the handler that do not have any corresponding annotations.
        
        Arguments:
            ann_handler {COCO_Annotation_Handler} -- [Reference COCO Annotation Handler]
            license_handler {COCO_License_Handler} -- [
                The COCO License Handler that you would like to update according to image removals.
                If None, licenses will not be updated.
            ]
        
        Keyword Arguments:
            id_list {List[int]} -- [
                The image IDs that you would like to check.
                If None, all images are checked.
            ] (default: {None})
        """
        rm_image_id_list = []
        if id_list is not None:
            for image_id in id_list:
                anns = ann_handler.get_annotations_from_imgIds([image_id])
                if len(anns) == 0:
                    rm_image_id_list.append(image_id)
        else:
            for coco_image in self:
                anns = ann_handler.get_annotations_from_imgIds([coco_image.id])
                if len(anns) == 0:
                    rm_image_id_list.append(coco_image.id)
        self.remove(rm_image_id_list, verbose=verbose)
        if license_handler is not None:
            pending_license_id_list = [license.id for license in license_handler]
            license_handler.remove_if_no_imgs(img_handler=self, id_list=pending_license_id_list, verbose=verbose)

class COCO_Annotation_Handler(BaseStructHandler['COCO_Annotation_Handler', 'COCO_Annotation']):
    def __init__(self, annotation_list: List[COCO_Annotation]=None):
        super().__init__(obj_type=COCO_Annotation, obj_list=annotation_list)
        self.annotation_list = self.obj_list

    def get_annotations_from_annIds(self, annIds: list) -> List[COCO_Annotation]:		
        return [ann for ann in self if ann.id in annIds]		
        
    def get_annotations_from_imgIds(self, imgIds: list) -> List[COCO_Annotation]:		
        return [ann for ann in self if ann.image_id in imgIds]
    
    def get_annotations_from_catIds(self, catIds: list) -> List[COCO_Annotation]:
        return [ann for ann in self if ann.category_id in catIds]

    def to_dict_list(self, strict: bool=True) -> List[dict]:
        return [item.to_dict(strict=strict) for item in self]

    def save_to_path(self, save_path: str, overwrite: bool=False, strict: bool=True):
        if file_exists(save_path) and not overwrite:
            logger.error(f'File already exists at save_path: {save_path}')
            raise Exception
        json_data = self.to_dict_list(strict=strict)
        json.dump(json_data, open(save_path, 'w'), indent=2, ensure_ascii=False)

    @classmethod
    def from_dict_list(cls, dict_list: List[dict], strict: bool=True) -> COCO_Annotation_Handler:
        return COCO_Annotation_Handler(
            annotation_list=[COCO_Annotation.from_dict(ann_dict, strict=strict) for ann_dict in dict_list]
        )

    @classmethod
    def load_from_path(cls, json_path: str, strict: bool=True) -> COCO_Annotation_Handler:
        check_file_exists(json_path)
        json_data = json.load(open(json_path, 'r'))
        return COCO_Annotation_Handler.from_dict_list(json_data, strict=strict)

    def remove(self, id_list: List[int], verbose: bool=False):
        # TODO: Create a base class that inherits from BaseStruct that requires an id class parameter
        # This method could be added to the base handler of the resulting object class.
        if len(id_list) > 0:
            idx_list = list(range(len(self)))
            idx_list.reverse()
            for idx in idx_list:
                if self[idx].id in id_list:
                    del self[idx]
                    if verbose:
                        logger.info(f'Deleted Annotation Id: {idx}')
    
    def remove_if_no_categories(
        self, cat_handler: COCO_Category_Handler,
        img_handler: COCO_Image_Handler=None, license_handler: COCO_License_Handler=None, id_list: List[int]=None, verbose: bool=False
    ):
        rm_ann_id_list = []
        pending_img_id_list = []
        existing_cat_id_list = [cat.id for cat in cat_handler]

        if id_list is not None:
            anns = self.get_annotations_from_annIds(id_list)
            for ann in anns:
                if ann.category_id not in existing_cat_id_list:
                    rm_ann_id_list.append(ann.id)
                    pending_img_id_list.append(ann.image_id)
        else:
            for ann in self:
                if ann.category_id not in existing_cat_id_list:
                    rm_ann_id_list.append(ann.id)
                    pending_img_id_list.append(ann.image_id)
        
        self.remove(rm_ann_id_list, verbose=verbose)
        if img_handler is not None:
            img_handler.remove_if_no_anns(ann_handler=self, license_handler=license_handler, id_list=pending_img_id_list, verbose=verbose)

class COCO_Category_Handler(BaseStructHandler['COCO_Category_Handler', 'COCO_Category']):
    def __init__(self, category_list: List[COCO_Category]=None):
        super().__init__(obj_type=COCO_Category, obj_list=category_list)
        self.category_list = self.obj_list

    def get_categories_from_name(self, name: str) -> List[COCO_Category]:		
        return [cat for cat in self if cat.name == name]

    def get_unique_category_from_name(self, name: str) -> COCO_Category:
        found_categories = self.get_categories_from_name(name)
        if len(found_categories) == 0:
            logger.error(f"Couldn't find any categories by the name: {name}")
            raise Exception
        elif len(found_categories) > 1:
            logger.error(f"Found {len(found_categories)} categories with the name {name}")
            logger.error(f"Found Categories:")
            for category in found_categories:
                logger.error(category)
            raise Exception
        return found_categories[0]

    def get_skeleton_from_name(self, name: str) -> (list, list):
        unique_category = self.get_unique_category_from_name(name)
        skeleton = unique_category.skeleton
        label_skeleton = unique_category.get_label_skeleton()
        return skeleton, label_skeleton

    def to_dict_list(self, strict: bool=True) -> List[dict]:
        return [item.to_dict(strict=strict) for item in self]

    def save_to_path(self, save_path: str, overwrite: bool=False, strict: bool=True):
        if file_exists(save_path) and not overwrite:
            logger.error(f'File already exists at save_path: {save_path}')
            raise Exception
        json_data = self.to_dict_list(strict=strict)
        json.dump(json_data, open(save_path, 'w'), indent=2, ensure_ascii=False)

    @classmethod
    def from_dict_list(cls, dict_list: List[dict], strict: bool=True) -> COCO_Category_Handler:
        return COCO_Category_Handler(
            category_list=[COCO_Category.from_dict(cat_dict, strict=strict) for cat_dict in dict_list]
        )

    @classmethod
    def load_from_path(cls, json_path: str, strict: bool=True) -> COCO_Category_Handler:
        check_file_exists(json_path)
        json_data = json.load(open(json_path, 'r'))
        return COCO_Category_Handler.from_dict_list(json_data, strict=strict)
    
    def remove(self, id_list: List[int], verbose: bool=False):
        # TODO: Create a base class that inherits from BaseStruct that requires an id class parameter
        # This method could be added to the base handler of the resulting object class.
        if len(id_list) > 0:
            idx_list = list(range(len(self)))
            idx_list.reverse()
            for idx in idx_list:
                if self[idx].id in id_list:
                    del self[idx]
                    if verbose:
                        logger.info(f'Deleted Category Id: {idx}')
    
    def remove_by_name(
        self, names: List[str],
        ann_handler: COCO_Annotation_Handler=None, img_handler: COCO_Image_Handler=None, license_handler: COCO_License_Handler=None,
        verbose: bool=False
    ):
        existing_category_names = [coco_cat.name for coco_cat in self]
        check_value_from_list(names, valid_value_list=existing_category_names)
        rm_ids = [coco_cat.id for coco_cat in self if coco_cat.name in names]
        self.remove(rm_ids, verbose=verbose)
        if ann_handler is not None:
            ann_handler.remove_if_no_categories(
                cat_handler=self,
                img_handler=img_handler,
                license_handler=license_handler,
                id_list=None,
                verbose=verbose
            )


from setuptools import setup, find_packages
import annotation_utils

packages = find_packages(
        where='.',
        include=['annotation_utils*']
)

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='pyclay-annotation_utils',
    version=annotation_utils.__version__,
    description='Utilities that are used for assisting with making annotated data.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/cm107/annotation_utils",
    author='Clayton Mork',
    author_email='mork.clayton3@gmail.com',
    license='MIT License',
    packages=packages,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'opencv-python>=4.1.1.26',
        'numpy>=1.17.2',
        'Shapely>=1.6.4.post2',
        'matplotlib>=3.1.1',
        'Pillow>=6.1.0',
        'pycocotools>=2.0.0',
        'pylint>=2.4.2',
        'labelme>=3.16.7',
        'PyYAML>=5.1.2',
        'pyqt5>=5.14.1',
        'pyclay-common_utils==0.1',
        'pyclay-logger==0.2'
    ],
    python_requires='>=3.7'
)


from logger import logger
from common_utils.file_utils import dir_exists, file_exists, move_file
from common_utils.path_utils import get_all_files_of_extension, get_rootname_from_path, \
    get_next_dump_path

class DataMover:
    def __init__(
        self, img_src: str, ann_src: str, img_dst: str, ann_dst: str,
        ann_extension: str, auto_renaming: bool=False, assume_labelme: bool=False
    ):
        self.img_src = img_src
        self.ann_src = ann_src
        self.img_dst = img_dst
        self.ann_dst = ann_dst
        self.ann_extension = ann_extension
        self.auto_renaming = auto_renaming
        self.assume_labelme = assume_labelme
        if not dir_exists(self.img_src):
            logger.error(f"Directory doesn't exist: {self.img_src}")
            raise Exception
        if not dir_exists(self.ann_src):
            logger.error(f"Directory doesn't exist: {self.ann_src}")
            raise Exception
        if not dir_exists(self.img_dst):
            logger.error(f"Directory doesn't exist: {self.img_dst}")
            raise Exception
        if not dir_exists(self.ann_dst):
            logger.error(f"Directory doesn't exist: {self.ann_dst}")
            raise Exception

    def get_pathlists(self) -> (list, list, list, list):
        img_src_pathlist = get_all_files_of_extension(dir_path=self.img_src, extension='png')
        ann_src_pathlist = get_all_files_of_extension(dir_path=self.ann_src, extension=self.ann_extension)
        img_dst_pathlist = get_all_files_of_extension(dir_path=self.img_dst, extension='png')
        ann_dst_pathlist = get_all_files_of_extension(dir_path=self.ann_dst, extension=self.ann_extension)
        return img_src_pathlist, ann_src_pathlist, img_dst_pathlist, ann_dst_pathlist

    def move(self):
        img_src_pathlist, ann_src_pathlist, img_dst_pathlist, ann_dst_pathlist = self.get_pathlists()
        nonempty = self.is_nonempty(self.img_src, img_src_pathlist, self.ann_src, ann_src_pathlist)
        if nonempty:
            self.check_dst_size(img_dst_pathlist, ann_dst_pathlist)
            self.check_all_ann_exists_in_img(img_src_pathlist, ann_src_pathlist, img_dst_pathlist, ann_dst_pathlist)
            if not self.assume_labelme:
                self.move_all_src_to_dst(img_src_pathlist, ann_src_pathlist)
            else:
                self.labelme_move_all_src_to_dst(img_dst_pathlist, ann_src_pathlist)
            self.post_move_check()
        else:
            logger.warning(f"Since the src directory is empty, there is nothing to be done.")

    def is_nonempty(self, img_dir: str, img_pathlist: list, ann_dir: str, ann_pathlist: list):
        if len(img_pathlist) == 0:
            logger.warning(f"{img_dir} is empty.")
            return False
        if len(ann_pathlist) == 0:
            logger.warning(f"{ann_dir} is empty.")
            return False
        return True

    def check_ann_empty(self, ann_dir: str, ann_pathlist):
        if len(ann_pathlist) > 0:
            logger.error(f"{ann_dir} is not empty.")
            raise Exception

    def check_dst_size(self, img_dst_pathlist: list, ann_dst_pathlist: list):
        if len(img_dst_pathlist) != len(ann_dst_pathlist):
            logger.error(f"Directory size mismatch.")
            logger.error(f"len(img_dst_pathlist) = {len(img_dst_pathlist)} != {len(ann_dst_pathlist)} = len(ann_dst_pathlist)")
            raise Exception

    def check_ann_exists_in_img(self, img_pathlist: list, ann_pathlist: list, img_dir_path: str):
        paths_not_found = []
        for ann_path in ann_pathlist:
            rootname = get_rootname_from_path(ann_path)
            corresponding_img_path = f"{img_dir_path}/{rootname}.png"
            if not file_exists(corresponding_img_path):
                paths_not_found.append(corresponding_img_path)
        if len(paths_not_found) > 0:
            for path_not_found in paths_not_found:
                logger.error(f"File not found: {path_not_found}")
            raise Exception

    def check_all_ann_exists_in_img(self, img_src_pathlist: list, ann_src_pathlist: list, img_dst_pathlist: list, ann_dst_pathlist: list):
        self.check_ann_exists_in_img(img_src_pathlist, ann_src_pathlist, self.img_src)
        self.check_ann_exists_in_img(img_dst_pathlist, ann_dst_pathlist, self.img_dst)

    def move_all_src_to_dst(self, img_src_pathlist: list, ann_src_pathlist: list):
        for ann_src_path in ann_src_pathlist:
            rootname = get_rootname_from_path(ann_src_path)
            img_src_path = f"{self.img_src}/{rootname}.png"
            if not self.auto_renaming:
                ann_dst_path = f"{self.ann_dst}/{rootname}.{self.ann_extension}"
                img_dst_path = f"{self.img_dst}/{rootname}.png"
                
            else:
                ann_dst_path = get_next_dump_path(
                    dump_dir=self.ann_dst,
                    file_extension=self.ann_extension
                )
                rootname = get_rootname_from_path(ann_dst_path)
                img_dst_path = f"{self.img_dst}/{rootname}.png"

            self.premove_check(img_src_path, img_dst_path, ann_src_path, ann_dst_path)
            move_file(ann_src_path, ann_dst_path, silent=False)
            move_file(img_src_path, img_dst_path, silent=False)

    def labelme_move_all_src_to_dst(self, img_src_pathlist: list, ann_src_pathlist: list):
        from .util.labelme_utils import move_annotation

        for ann_src_path in ann_src_pathlist:
            rootname = get_rootname_from_path(ann_src_path)
            img_src_path = f"{self.img_src}/{rootname}.png"
            if not self.auto_renaming:
                ann_dst_path = f"{self.ann_dst}/{rootname}.{self.ann_extension}"
                img_dst_path = f"{self.img_dst}/{rootname}.png"
                
            else:
                ann_dst_path = get_next_dump_path(
                    dump_dir=self.ann_dst,
                    file_extension=self.ann_extension
                )
                rootname = get_rootname_from_path(ann_dst_path)
                img_dst_path = f"{self.img_dst}/{rootname}.png"

            self.premove_check(img_src_path, img_dst_path, ann_src_path, ann_dst_path)
            move_file(img_src_path, img_dst_path, silent=False)
            move_annotation(
                src_img_path=img_src_path,
                src_json_path=ann_src_path,
                dst_img_path=img_dst_path,
                dst_json_path=ann_dst_path,
                bound_type='rect'
            )
            src_preview = '/'.join(ann_src_path.split('/')[-3:])
            dest_preview = '/'.join(ann_dst_path.split('/')[-3:])
            print('Moved {} to {}'.format(src_preview, dest_preview))

    def premove_check(self, img_src_path: str, img_dst_path: str, ann_src_path: str, ann_dst_path: str):
        paths_not_found = []
        paths_already_exists = []
        if not file_exists(img_src_path):
            paths_not_found.append(img_src_path)
        if file_exists(img_dst_path):
            paths_already_exists.append(img_dst_path)
        if not file_exists(ann_src_path):
            paths_not_found.append(ann_src_path)
        if file_exists(ann_dst_path):
            paths_already_exists.append(ann_dst_path)
        if len(paths_not_found) > 0:
            for path_not_found in paths_not_found:
                logger.error(f"File not found: {path_not_found}")
            raise Exception
        if len(paths_already_exists) > 0:
            for path_already_exists in paths_already_exists:
                logger.error(f"File already exists: {path_already_exists}")
            raise Exception

    def post_move_check(self):
        img_src_pathlist, ann_src_pathlist, img_dst_pathlist, ann_dst_pathlist = self.get_pathlists()
        nonempty = self.is_nonempty(self.img_dst, img_dst_pathlist, self.ann_dst, ann_dst_pathlist)
        if nonempty:
            self.check_ann_empty(self.ann_src, ann_src_pathlist)
            logger.info(f"There are currently {len(img_src_pathlist)} unlabeled images in {self.img_src}")
        else:
            logger.error(f"The destination directories is empty after post-move. Something has gone wrong.")
            raise Exception



import yaml
import json
from logger import logger
from common_utils.check_utils import check_type_from_list, check_type, check_value_from_list, \
    check_file_exists, check_value
from common_utils.path_utils import get_extension_from_path

class DatasetPathConfig:
    def __init__(self):
        self.data = None

        self.valid_extensions = ['json', 'yaml']
        self.main_required_keys = ['collection_dir', 'dataset_names', 'dataset_specific']
        self.specific_required_keys = ['img_dir', 'ann_path', 'ann_format']
        self.valid_ann_formats = ['coco', 'custom']

    @classmethod
    def from_load(self, target, verbose: bool=False):
        config = DatasetPathConfig()
        config.load(target=target, verbose=verbose)
        return config

    def check_valid_config(self, collection_dict_list: list):
        check_type(item=collection_dict_list, valid_type_list=[list])
        for i, collection_dict in enumerate(collection_dict_list):
            check_type(item=collection_dict, valid_type_list=[dict])
            check_value_from_list(
                item_list=list(collection_dict.keys()),
                valid_value_list=self.main_required_keys
            )
            for required_key in self.main_required_keys:
                if required_key not in collection_dict.keys():
                    logger.error(f"collection_dict at index {i} is missing required key: {required_key}")
                    raise Exception
            collection_dir = collection_dict['collection_dir']
            check_type(item=collection_dir, valid_type_list=[str])
            dataset_names = collection_dict['dataset_names']
            check_type(item=dataset_names, valid_type_list=[list])
            check_type_from_list(item_list=dataset_names, valid_type_list=[str])
            dataset_specific = collection_dict['dataset_specific']
            check_type(item=dataset_specific, valid_type_list=[dict])
            check_value_from_list(
                item_list=list(dataset_specific.keys()),
                valid_value_list=self.specific_required_keys
            )
            for required_key in self.specific_required_keys:
                if required_key not in dataset_specific.keys():
                    logger.error(f"dataset_specific at index {i} is missing required key: {required_key}")
                    raise Exception
            img_dir = dataset_specific['img_dir']
            ann_path = dataset_specific['ann_path']
            ann_format = dataset_specific['ann_format']
            check_type_from_list(item_list=[img_dir, ann_path, ann_format], valid_type_list=[str, list])
            if type(img_dir) is list and len(img_dir) != len(dataset_names):
                logger.error(f"Length mismatch at index: {i}")
                logger.error(f"type(img_dir) is list but len(img_dir) == {len(img_dir)} != {len(dataset_names)} == len(dataset_names)")
                raise Exception
            if type(ann_path) is list and len(ann_path) != len(dataset_names):
                logger.error(f"Length mismatch at index: {i}")
                logger.error(f"type(ann_path) is list but len(ann_path) == {len(ann_path)} != {len(dataset_names)} == len(dataset_names)")
                raise Exception
            if type(ann_format) is list and len(ann_format) != len(dataset_names):
                logger.error(f"Length mismatch at index: {i}")
                logger.error(f"type(ann_format) is list but len(ann_format) == {len(ann_format)} != {len(dataset_names)} == len(dataset_names)")
                raise Exception

            if type(ann_format) is str:
                check_value(item=ann_format, valid_value_list=self.valid_ann_formats)
            elif type(ann_format) is list:
                check_value_from_list(item_list=ann_format, valid_value_list=self.valid_ann_formats)
            else:
                raise Exception

    def write_config(self, dest_path: str, verbose: bool=False):
        extension = get_extension_from_path(dest_path)
        if extension == 'yaml':
            yaml.dump(self.data, open(dest_path, 'w'), allow_unicode=True)
        elif extension == 'json':
            json.dump(self.data, open(dest_path, 'w'), indent=2, ensure_ascii=False)
        else:
            logger.error(f"Invalid extension: {extension}")
            logger.error(f"Valid file extensions: {self.valid_extensions}")
            raise Exception
        if verbose:
            logger.good(f"Dataset path config written successfully to:\n{dest_path}")
    
    def load(self, target, verbose: bool=False):
        if type(target) is str:
            extension = get_extension_from_path(target)
            if extension == 'yaml':
                check_file_exists(target)
                loaded_data = yaml.load(open(target, 'r'), Loader=yaml.FullLoader)
            elif extension == 'json':
                check_file_exists(target)
                loaded_data = json.load(open(target, 'r'))
            else:
                logger.error(f"Invalid extension: {extension}")
                logger.error(f"Note that string targets are assumed to be paths.")
                logger.error(f"Valid file extensions: {self.valid_extensions}")
                raise Exception
        elif type(target) is list:
            loaded_data = target
        elif type(target) is dict:
            loaded_data = [target]
        else:
            logger.error(f"Invalid target type: {type(target)}")
            raise Exception
        self.check_valid_config(collection_dict_list=loaded_data)
        self.data = loaded_data
        if verbose:
            logger.good(f"Dataset path config has been loaded successfully.")

    def _get_img_dir_list(self, dataset_paths: list, img_dir) -> list:
        check_type(item=img_dir, valid_type_list=[str, list])
        if type(img_dir) is str:
            img_dir_list = [f"{dataset_path}/{img_dir}" for dataset_path in dataset_paths]
        elif type(img_dir) is list:
            if len(img_dir) == len(dataset_paths):
                check_type_from_list(item_list=img_dir, valid_type_list=[str])
                img_dir_list = [f"{dataset_path}/{img_dir_path}" for dataset_path, img_dir_path in zip(dataset_paths, img_dir)]
            else:
                logger.error(f"type(img_dir) is list but len(img_dir) == {len(img_dir)} != {len(dataset_paths)} == len(dataset_paths)")
                raise Exception
        else:
            raise Exception
        return img_dir_list

    def _get_ann_path_list(self, dataset_paths: list, ann_path) -> list:
        check_type(item=ann_path, valid_type_list=[str, list])
        if type(ann_path) is str:
            ann_path_list = [f"{dataset_path}/{ann_path}" for dataset_path in dataset_paths]
        elif type(ann_path) is list:
            if len(ann_path) == len(dataset_paths):
                check_type_from_list(item_list=ann_path, valid_type_list=[str])
                ann_path_list = [f"{dataset_path}/{ann_path_path}" for dataset_path, ann_path_path in zip(dataset_paths, ann_path)]
            else:
                logger.error(f"type(ann_path) is list but len(ann_path) == {len(ann_path)} != {len(dataset_paths)} == len(dataset_paths)")
                raise Exception
        else:
            raise Exception
        return ann_path_list

    def _get_ann_format_list(self, dataset_paths: list, ann_format) -> list:
        check_type(item=ann_format, valid_type_list=[str, list])
        
        if type(ann_format) is str:
            check_value(item=ann_format, valid_value_list=self.valid_ann_formats)
            ann_format_list = [ann_format] * len(dataset_paths)
        elif type(ann_format) is list:
            check_value_from_list(item_list=ann_format, valid_value_list=self.valid_ann_formats)
            if len(ann_format) == len(dataset_paths):
                check_type_from_list(item_list=ann_format, valid_type_list=[str])
                ann_format_list = ann_format
            else:
                logger.error(f"type(ann_format) is list but len(ann_format) == {len(ann_format)} != {len(dataset_paths)} == len(dataset_paths)")
                raise Exception
        else:
            raise Exception
        return ann_format_list

    def get_paths(self) -> (list, list, list, list):
        if self.data is None:
            logger.error(f"Config hasn't been loaded yet. Please use load() method.")
            raise Exception

        combined_dataset_dir_list = []
        combined_img_dir_list = []
        combined_ann_path_list = []
        combined_ann_format_list = []
        for collection_dict in self.data:
            collection_dir = collection_dict['collection_dir']
            dataset_names = collection_dict['dataset_names']
            img_dir = collection_dict['dataset_specific']['img_dir']
            ann_path = collection_dict['dataset_specific']['ann_path']
            ann_format = collection_dict['dataset_specific']['ann_format']
            dataset_paths = [f'{collection_dir}/{dataset_name}' for dataset_name in dataset_names]
            img_dir_list = self._get_img_dir_list(dataset_paths=dataset_paths, img_dir=img_dir)
            ann_path_list = self._get_ann_path_list(dataset_paths=dataset_paths, ann_path=ann_path)
            ann_format_list = self._get_ann_format_list(dataset_paths=dataset_paths, ann_format=ann_format)
            combined_dataset_dir_list.extend(dataset_paths)
            combined_img_dir_list.extend(img_dir_list)
            combined_ann_path_list.extend(ann_path_list)
            combined_ann_format_list.extend(ann_format_list)

        return combined_dataset_dir_list, combined_img_dir_list, combined_ann_path_list, combined_ann_format_list


from __future__ import annotations
from logger import logger
from common_utils.time_utils import get_present_year, get_present_time_Ymd
from common_utils.user_utils import get_username
from common_utils.path_utils import get_filename

from ....coco.structs import COCO_Info, \
    COCO_License, COCO_Image, COCO_Annotation, COCO_Category, \
    COCO_License_Handler, COCO_Image_Handler, \
    COCO_Annotation_Handler, COCO_Category_Handler
from ....coco.annotation.coco_annotation import COCO_AnnotationFileParser
from .id_map import COCO_Mapper_Handler

class COCO_Field_Buffer:
    def __init__(
        self, info: COCO_Info, licenses: COCO_License_Handler, images: COCO_Image_Handler,
        annotations: COCO_Annotation_Handler, categories: COCO_Category_Handler
    ):
        self.info = info
        self.licenses = licenses
        self.images = images
        self.annotations = annotations
        self.categories = categories

    @classmethod
    def from_scratch(self, description: str, url: str, version: str='1.0') -> COCO_Field_Buffer:
        info = COCO_Info(
            description=description,
            url=url,
            version=version,
            year=get_present_year(),
            contributor=get_username(),
            date_created=get_present_time_Ymd()
        )
        licenses = COCO_License_Handler()
        images = COCO_Image_Handler()
        annotations = COCO_Annotation_Handler()
        categories = COCO_Category_Handler()
        return COCO_Field_Buffer(
            info=info, licenses=licenses,
            images=images, annotations=annotations,
            categories=categories
        )

    @classmethod
    def from_parser(
        self, parser: COCO_AnnotationFileParser,
        new_description: str=None, new_url: str=None, new_version: str='1.0'
    ) -> COCO_Field_Buffer:
        info = parser.info
        info.description = new_description if new_description is not None else info.description
        info.url = new_url if new_url is not None else info.url
        info.version = new_version if new_version is not None else info.version

        return COCO_Field_Buffer(
            info=info,
            licenses=parser.licenses,
            images=parser.images,
            annotations=parser.annotations,
            categories=parser.categories
        )

    def update_info(self, description: str, url: str, version: str='1.0'):
        self.info = COCO_Info(
            description=description,
            url=url,
            version=version,
            year=self.info.year,
            contributor=self.info.contributor,
            date_created=self.info.date_created
        )

    def update_fields(
        self, licenses: COCO_License_Handler, images: COCO_Image_Handler,
        annotations: COCO_Annotation_Handler, categories: COCO_Category_Handler,
        info: COCO_Info=None
    ):
        self.licenses = licenses
        self.images = images
        self.annotations = annotations
        self.categories = categories
        if info is not None:
            self.info = info

    def process_license(self, coco_license: COCO_License, id_mapper: COCO_Mapper_Handler, unique_key: str) -> (bool, int, int):
        pending_coco_license = coco_license.copy()

        found = False
        added_new = False
        found_coco_license = None
        for loaded_coco_license in self.licenses.license_list:
            if pending_coco_license.url == loaded_coco_license.url and pending_coco_license.name == loaded_coco_license.name:
                found = True
                found_coco_license = loaded_coco_license
                break
        old_id = pending_coco_license.id
        if not found:
            new_id = len(self.licenses.license_list)
            pending_coco_license.id = new_id
            self.licenses.add(pending_coco_license)
            added_new = True
        else: # Add map only
            new_id = found_coco_license.id

        id_mapper.license_mapper.add(
            unique_key=unique_key,
            old_id=old_id,
            new_id=new_id
        )
        return added_new, old_id, new_id

    def process_image(self, coco_image: COCO_Image, id_mapper: COCO_Mapper_Handler, unique_key: str, img_dir: str=None, update_img_path: bool=False) -> (bool, int, int):
        pending_coco_image = coco_image.copy()

        if update_img_path:
            if img_dir is None:
                logger.error(f"img_dir is required in order to update the img_path")
                raise Exception
            coco_url_filename = get_filename(pending_coco_image.coco_url)
            if coco_url_filename != pending_coco_image.file_name:
                logger.error(f"coco_url_filename == {coco_url_filename} != pending_coco_image.file_name == {pending_coco_image.file_name}")
                raise Exception
            pending_coco_image.coco_url = f"{img_dir}/{coco_url_filename}"

        # No checks should be necessary
        old_id = pending_coco_image.id
        new_id = len(self.images.image_list)

        # Get License Id
        found, new_license_id = id_mapper.license_mapper.get_new_id(
            unique_key=unique_key, old_id=pending_coco_image.license_id
        )
        if not found:
            logger.error(f"Couldn't find unique_key: {unique_key}, old_id: {pending_coco_image.license_id} pair in license_mapper.")
            raise Exception

        # Update Id References
        pending_coco_image.license_id = new_license_id
        pending_coco_image.id = new_id
        
        self.images.add(pending_coco_image)
        added_new = True
        id_mapper.image_mapper.add(
            unique_key=unique_key,
            old_id=old_id,
            new_id=new_id
        )
        return added_new, old_id, new_id

    def process_annotation(self, coco_annotation: COCO_Annotation, id_mapper: COCO_Mapper_Handler, unique_key: str) -> (bool, int, int):
        pending_coco_annotation = coco_annotation.copy()
        
        # No checks should be necessary
        old_id = pending_coco_annotation.id
        new_id = len(self.annotations.annotation_list)
        
        # Get Image Id
        found, new_image_id = id_mapper.image_mapper.get_new_id(
            unique_key=unique_key, old_id=pending_coco_annotation.image_id
        )
        if not found:
            logger.error(f"Couldn't find unique_key: {unique_key}, old_id: {pending_coco_annotation.image_id} pair in image_mapper.")
            raise Exception

        # Get Category Id
        found, new_category_id = id_mapper.category_mapper.get_new_id(
            unique_key=unique_key, old_id=pending_coco_annotation.category_id
        )
        if not found:
            logger.error(f"Couldn't find unique_key: {unique_key}, old_id: {pending_coco_annotation.category_id} pair in category_mapper.")
            raise Exception

        # Update Id References
        pending_coco_annotation.image_id = new_image_id
        pending_coco_annotation.category_id = new_category_id
        pending_coco_annotation.id = new_id
        
        self.annotations.add(pending_coco_annotation)
        added_new = True
        id_mapper.annotation_mapper.add(
            unique_key=unique_key,
            old_id=old_id,
            new_id=new_id
        )
        return added_new, old_id, new_id

    def process_category(self, coco_category: COCO_Category, id_mapper: COCO_Mapper_Handler, unique_key: str) -> (bool, int, int):
        pending_coco_category = coco_category.copy()
        
        found = False
        added_new = False
        found_coco_category = None
        for loaded_coco_category in self.categories.category_list:
            if pending_coco_category.supercategory == loaded_coco_category.supercategory and \
                pending_coco_category.name == loaded_coco_category.name:
                found = True
                found_coco_category = loaded_coco_category
                break
        old_id = pending_coco_category.id
        if not found:
            new_id = len(self.categories.category_list)
            pending_coco_category.id = new_id
            self.categories.add(pending_coco_category)
            added_new = True
        else: # Add map only
            new_id = found_coco_category.id

        id_mapper.category_mapper.add(
            unique_key=unique_key,
            old_id=old_id,
            new_id=new_id
        )
        return added_new, old_id, new_id